// ############################################################################################
//  vdspenv.cpp
//  Implementation of _CVortexAudioEnv class using direct access to /dev/dsp
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <auddsp/vdspsf.h>
#include <auddsp/vdspenv.h>

// ============================================================================================
// Exported variables declaration
// ============================================================================================
// ***** The one and the only DSP environment instance
_CVortexAudioEnv *_pcsAudioEnv_ = NULL;

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The audio environment class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexAudioEnv::_CVortexAudioEnv()
{
  if(_pcsAudioEnv_ == NULL) _pcsAudioEnv_ = this;
  mt_tD.n32DSPFD = -1; 
  mt_tD.pucPBuff = NULL;
  memset(mt_atPSB, 0, sizeof(mt_atPSB[0]) * _MAX_SB_);
}
_CVortexAudioEnv::~_CVortexAudioEnv()
{
  // ..... Temporary variables
  _vint32 n32Dummy;
  
  // ..... Terminate threads
  if(!mt_csTWriter.Stop()) mt_csTWriter.Kill();

  // ..... Free memory
  if(mt_tD.pucPBuff) free(mt_tD.pucPBuff);

  // ..... Close audio device
  if(mt_tD.n32DSPFD >= 0)
  {
    ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_RESET, &n32Dummy);
    close(mt_tD.n32DSPFD);
  }
  
  // ..... Set reference to NULL
  _pcsAudioEnv_ = NULL;
}

// ***** Initialization functions
_vbool _CVortexAudioEnv::QueryDeviceName(_vchar *pcResult)
{
  // ..... Temporary variables
  struct stat tStat;
  _vuint8     un8Cnt;
  _vchar      acTmp[64];
  
  // ..... Check
  if(_pcsAudioEnv_ != this)
  {
    VSetErrorCode(VERR_MultipleInstance);
    VError(_F_, _L_, "Cannot have multiple instance of _CVortexAudioEnv");
    return(false);
  }

  // ..... Search for /dev/sound/dsp*
  pcResult[0] = 0;
  if((stat("/dev/sound", &tStat) == 0) && S_ISDIR(tStat.st_mode))
  {
    if((stat("/dev/sound/dsp", &tStat) == 0) && S_ISCHR(tStat.st_mode))
    {
      strcpy(pcResult, "/dev/sound/dsp");
      return(true);
    }
    else
    {
      for(un8Cnt = 0; un8Cnt < 128; un8Cnt++)
      {
        sprintf(acTmp, "/dev/sound/dsp%d", un8Cnt);
        if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode))
        {
          strcpy(pcResult, acTmp);
          return(true);
        }
      }
    }
  }

  // ..... Search for /dev/dsp*
  if((stat("/dev/dsp", &tStat) == 0) && S_ISCHR(tStat.st_mode))
  {
    strcpy(pcResult, "/dev/dsp");
    return(true);
  }
  else
  {
    for(un8Cnt = 0; un8Cnt < 128; un8Cnt++)
    {
      sprintf(acTmp, "/dev/dsp%d", un8Cnt);
      if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode))
      {
        strcpy(pcResult, acTmp);
        return(true);
      }
    }
  }

  // ..... Search for /dev/audio*
  if((stat("/dev/audio", &tStat) == 0) && S_ISCHR(tStat.st_mode))
  {
    strcpy(pcResult, "/dev/audio");
    return(true);
  }
  else
  {
    for(un8Cnt = 0; un8Cnt < 128; un8Cnt++)
    {
      sprintf(acTmp, "/dev/audio%d", un8Cnt);
      if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode))
      {
        strcpy(pcResult, acTmp);
        return(true);
      }
    }
  }
  
  // ..... Not found
  VSetErrorCode(VERR_NotFound);
  VError(_F_, _L_, "Could not found any audio device in this system");
  return(false);
}
_vbool _CVortexAudioEnv::OpenDevice(_vchar *pcDeviceName)
{
  // ..... Temporary variables
  _vint32 n32Dummy;
  
  // ..... Check
  if(_pcsAudioEnv_ != this)
  {
    VSetErrorCode(VERR_MultipleInstance);
    VError(_F_, _L_, "Cannot have multiple instance of _CVortexAudioEnv");
    return(false);
  }

  // ..... Open device
  mt_tD.n32DSPFD = open(pcDeviceName, O_WRONLY);
  if(mt_tD.n32DSPFD < 0)
  {
    VSetErrorCode(VERR_OpenDevice);
    VError(_F_, _L_, "Could not open audio device \"%s\"", pcDeviceName);
    return(false);
  }
  ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_RESET, &n32Dummy);
  strcpy(mt_tD.acDevName, pcDeviceName);
  
  // ..... OK !
  return(true);
}
_vbool _CVortexAudioEnv::InitializeDevice(const _TVortexAudSettings *ptPrefered, _TVortexAudSettings *ptRealized)
{
  // ..... Temporary variables
  _TVortexAudSettings tR, tP = { AFMT_S16_LE, 44100, 2 };
  _TVortexAudSettings *ptR, *ptP;
  _vint32             n32Tmp, n32Frag;
  _vchar              *pcTmp;
  
  // ..... Check
  if(_pcsAudioEnv_ != this)
  {
    VSetErrorCode(VERR_MultipleInstance);
    VError(_F_, _L_, "Cannot have multiple instance of _CVortexAudioEnv");
    return(false);
  }
  if(mt_tD.n32DSPFD < 0)
  {
    VSetErrorCode(VERR_DeviceNotOpenend);
    VError(_F_, _L_, "Audio device is not opened");
    return(false);
  }

  // ..... Preparations
  if(ptPrefered) 
    ptP = (_TVortexAudSettings *) ptPrefered;
  else
    ptP = &tP;
  if(ptRealized) 
    ptR = ptRealized;
  else
    ptR = &tR;
  
  // ..... Get supported audio formats
  if(ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_GETFMTS, &n32Tmp) < 0)
  {
    VSetErrorCode(VERR_GetDeviceParam);
    VError(_F_, _L_, "Failed to get supported audio formats");
    return(false);
  }
  
  // ..... Find usable audio format
  if((n32Tmp & ptP->n32Format) != ptP->n32Format)
  {
    if(n32Tmp & AFMT_S16_LE)
      ptR->n32Format = AFMT_S16_LE;
    else if(n32Tmp & AFMT_U16_LE)
      ptR->n32Format = AFMT_U16_LE;
    else if(n32Tmp & AFMT_S8)
      ptR->n32Format = AFMT_S8;
    else if(n32Tmp & AFMT_U8)
      ptR->n32Format = AFMT_U8;
    else if(n32Tmp & AFMT_S16_BE)
      ptR->n32Format = AFMT_S16_BE;
    else if(n32Tmp & AFMT_U16_BE)
      ptR->n32Format = AFMT_U16_BE;
    if(ptR->n32Format == 0)
    {
      VSetErrorCode(VERR_NotSupported);
      VError(_F_, _L_, "Could not find any usable audio format");
      return(false);
    }
  }
  else
    ptR->n32Format = ptP->n32Format;
  
  // ..... Find usable channels
  n32Tmp = ptP->n32Channels;
  if(n32Tmp < 1)
  {
    n32Tmp = 1;
    VWarning(_F_, _L_, "Only 1 channel or 2 channels audio supported by _CVortexAudioEnv");
  }
  if(n32Tmp > 2)
  {
    n32Tmp = 2;
    VWarning(_F_, _L_, "Only 1 channel or 2 channels audio supported by _CVortexAudioEnv");
  }
  if(ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_CHANNELS, &n32Tmp) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio channels");
    return(false);
  }
  ptR->n32Channels = n32Tmp;
  
  // ..... Find usable frequency
  n32Tmp = ptP->n32Frequency;
  if(ioctl(mt_tD.n32DSPFD, SOUND_PCM_WRITE_RATE, &n32Tmp) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio frequency (sampling rate)");
    return(false);
  }
  ptR->n32Frequency = n32Tmp;

  // ..... Compare the prefered and the realized settings
  if(ptPrefered) 
  {
    if(memcmp(ptR, ptP, sizeof(tR)) != 0)
      VWarning(_F_, _L_, "The realized DSP settings are different than the prefered settings");
  }
  
  // ..... Calculate buffer size
  //       The bigger the buffer the better it perfoms on slow computer, however,
  //       bigger buffer will cause more delay when playing sound
  if((ptR->n32Format == AFMT_U8) || (ptR->n32Format == AFMT_S8))
    n32Tmp = 2 * 1024 * ptR->n32Channels;
  else
    n32Tmp = 2 * 1024 * ptR->n32Channels * 2;

  // ..... Calculate fragment size
  for(n32Frag = 0; (0x01 << n32Frag) < n32Tmp; n32Frag++);
  n32Frag |= 0x00020000;
  
  // ..... Reopen audio device
  close(mt_tD.n32DSPFD);
  mt_tD.n32DSPFD = open(mt_tD.acDevName, O_WRONLY);
  if(mt_tD.n32DSPFD < 0)
  {
    VSetErrorCode(VERR_OpenDevice);
    VError(_F_, _L_, "Failed to reopen audio device \"%s\"", mt_tD.acDevName);
    return(false);
  }
  ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_RESET, &n32Tmp);
  
  // ..... Set fixed DSP settings
  if(ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_SETFMT, &ptR->n32Format) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio format");
    return(false);
  }
  if(ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_SETFRAGMENT, &n32Frag) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio fragment");
    return(false);
  }
  if(ioctl(mt_tD.n32DSPFD, SNDCTL_DSP_CHANNELS, &ptR->n32Channels) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio channels");
    return(false);
  }
  if(ioctl(mt_tD.n32DSPFD, SOUND_PCM_WRITE_RATE, &ptR->n32Frequency) < 0)
  {
    VSetErrorCode(VERR_SetDeviceParam);
    VError(_F_, _L_, "Failed to set audio frequency (sampling rate)");
    return(false);
  }
  
  // ..... Copy formats
  if(ptR->n32Format == AFMT_U8)
  {
    mt_tD.tFormat.un8Bits   = 8;
    mt_tD.tFormat.un8Sign   = 0;
    mt_tD.tFormat.un8Endian = 0;
  }
  else if(ptR->n32Format == AFMT_S8)
  {
    mt_tD.tFormat.un8Bits   = 8;
    mt_tD.tFormat.un8Sign   = 1;
    mt_tD.tFormat.un8Endian = 0;
  }
  else if(ptR->n32Format == AFMT_U16_LE)
  {
    mt_tD.tFormat.un8Bits   = 16;
    mt_tD.tFormat.un8Sign   = 0;
    mt_tD.tFormat.un8Endian = 0;
  }
  else if(ptR->n32Format == AFMT_U16_BE)
  {
    mt_tD.tFormat.un8Bits   = 16;
    mt_tD.tFormat.un8Sign   = 0;
    mt_tD.tFormat.un8Endian = 1;
  }
  else if(ptR->n32Format == AFMT_S16_LE)
  {
    mt_tD.tFormat.un8Bits   = 16;
    mt_tD.tFormat.un8Sign   = 1;
    mt_tD.tFormat.un8Endian = 0;
  }
  else if(ptR->n32Format == AFMT_S16_BE)
  {
    mt_tD.tFormat.un8Bits   = 16;
    mt_tD.tFormat.un8Sign   = 1;
    mt_tD.tFormat.un8Endian = 1;
  }
  mt_tD.tFormat.un8Channels   = ptR->n32Channels;
  mt_tD.tFormat.un32Frequency = ptR->n32Frequency;
  mt_tD.un32FragSize  = 1 << (n32Frag & 0x0000FFFF);
  mt_tD.un32FragSize2 = mt_tD.un32FragSize >> 1;
  mt_tD.un32FragSize4 = mt_tD.un32FragSize >> 2;
  mt_tD.un32PBuffSize = mt_tD.un32FragSize * 4;
  
  // ..... Allocate memory for primary buffer
  mt_tD.pucPBuff = (_vuchar *) malloc(mt_tD.un32PBuffSize);
  if(mt_tD.pucPBuff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for primary sound buffer");
    return(false);
  }
  mt_tD.pucPBuffRead  = mt_tD.pucPBuff;
  mt_tD.pucPBuffWrite = mt_tD.pucPBuff;
  mt_tD.pucPBuffMax   = mt_tD.pucPBuff + mt_tD.un32PBuffSize - mt_tD.un32FragSize;
  mt_tD.un8ToWrite    = 0;
  
  // ..... Initialize threads
  if(!mt_csTWriter.Initialize(_DSPWriter_, this)) return(false);

  // ..... Initialize sound mixing function
  //       Assuming that this system is a little-endian (x86)
  if(mt_tD.tFormat.un8Channels == 1)
  {
    if(mt_tD.tFormat.un8Bits == 8)
    {
      // U8 mono
      if(mt_tD.tFormat.un8Sign == 0)
        mt_pfnDoSoundMixing = _DSPMixerU8m_;
      // S8 mono
      else
        mt_pfnDoSoundMixing = _DSPMixerS8m_;
    }
    else
    {
      if(mt_tD.tFormat.un8Sign == 0)
      {
        // U16 LE mono
        if(mt_tD.tFormat.un8Endian == 0)
          mt_pfnDoSoundMixing = _DSPMixerU16m_;
        // U16 BE mono
        else
          mt_pfnDoSoundMixing = _DSPIMixerU16m_;
      }
      else
      {
        // S16 LE mono
        if(mt_tD.tFormat.un8Endian == 0)
          mt_pfnDoSoundMixing = _DSPMixerS16m_;
        // S16 BE mono
        else
          mt_pfnDoSoundMixing = _DSPIMixerS16m_;
      }
    }
  }
  else
  {
    if(mt_tD.tFormat.un8Bits == 8)
    {
      // U8 stereo
      if(mt_tD.tFormat.un8Sign == 0)
        mt_pfnDoSoundMixing = _DSPMixerU8s_;
      // S8 stereo
      else
        mt_pfnDoSoundMixing = _DSPMixerS8s_;
    }
    else
    {
      if(mt_tD.tFormat.un8Sign == 0)
      {
        // U16 LE stereo
        if(mt_tD.tFormat.un8Endian == 0)
          mt_pfnDoSoundMixing = _DSPMixerU16s_;
        // U16 BE stereo
        else
          mt_pfnDoSoundMixing = _DSPIMixerU16s_;
      }
      else
      {
        // S16 LE stereo
        if(mt_tD.tFormat.un8Endian == 0)
          mt_pfnDoSoundMixing = _DSPMixerS16s_;
        // S16 BE stereo
        else
          mt_pfnDoSoundMixing = _DSPIMixerS16s_;
      }
    }
  }
  
  // ..... Show info
  if(ptR->n32Format == AFMT_U8)
    pcTmp = "AFMT_U8";
  else if(ptR->n32Format == AFMT_S8)
    pcTmp = "AFMT_S8";
  else if(ptR->n32Format == AFMT_U16_LE)
    pcTmp = "AFMT_U16_LE";
  else if(ptR->n32Format == AFMT_U16_BE)
    pcTmp = "AFMT_U16_BE";
  else if(ptR->n32Format == AFMT_S16_LE)
    pcTmp = "AFMT_S16_LE";
  else
    pcTmp = "AFMT_S16_BE";
  VInfo(_F_, _L_, "Realized audio device settings :\n  Device path         = %s\n  Format              = %s\n  Fragments           = %d\n  Fragment size       = %u bytes\n  Channels            = %d\n  Frequency           = %d Hz\n  Primary buffer size = %u", 
        mt_tD.acDevName, pcTmp,
        n32Frag >> 16, mt_tD.un32FragSize,
        ptR->n32Channels , ptR->n32Frequency, mt_tD.un32PBuffSize);

  // ..... OK
  return(true);
}
_vbool _CVortexAudioEnv::ActivateDevice(_vvoid)
{
  // ..... Check
  if(mt_tD.n32DSPFD < 0)
  {
    VSetErrorCode(VERR_DeviceNotOpenend);
    VError(_F_, _L_, "Audio device is not opened");
    return(false);
  }

  // ..... Clear buffer
  _ClearBuffer_(mt_tD.pucPBuff, mt_tD.un32PBuffSize, mt_tD.tFormat.un8Sign, mt_tD.tFormat.un8Bits);

  // ..... Activate writer thread
  if(!mt_csTWriter.Start()) return(false);
  
  // ..... OK !
  VInfo(_F_, _L_, "DSP thread activated");
  return(true);
}
_vvoid _CVortexAudioEnv::DeactivateDevice(_vvoid)
{
  // ..... Check
  if(mt_tD.n32DSPFD < 0) return;

  // ..... Deactivate writer thread
  if(!mt_csTWriter.Stop()) mt_csTWriter.Kill();
  VInfo(_F_, _L_, "DSP thread deactivated");
}

// ***** Sound buffer functions
_vvoid _CVortexAudioEnv::_ClearBuffer_(_vuchar *pucBuff, _vuint32 un32BuffSize, _vuint8 un8Sign, _vuint8 un8Bits)
{
  // ..... Temporary variables
  _vuchar  *pucTmp = pucBuff;
  _vuint32 un32Cnt;

  // ..... Clear buffer
  //       Assuming that this system is a little-endian (x86)
  if(un8Sign)
    memset(pucBuff, 0x00, un32BuffSize);
  else
  {
    if(un8Bits == 8)
      memset(pucBuff, 0x80, un32BuffSize);
    else
    {
      for(un32Cnt = 0; un32Cnt < (un32BuffSize >> 1); un32Cnt++)
      {
        *pucTmp++ = 0x00;
        *pucTmp++ = 0x80;
      }
    }
  }
}

// ============================================================================================
// End of vdspenv.cpp
// ============================================================================================
