// ############################################################################################
//  vfataleh.cpp
//  Implementation of fatal signal handler functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <signal.h>
#include <stdlib.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <vfataleh.h>

// ============================================================================================
// Local variables declaration
// ============================================================================================
// ***** Used by signal handler
static sighandler_t       _atOldHandler_[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
static PFNVortexFatalFunc _pfnUser_         = NULL;
static _vvoid             *_pvUserData_     = NULL;
static _vbool             _bInstalled_      = false;
  
// ============================================================================================
// Local functions prototypes
// ============================================================================================
// ***** Handlers
static _vvoid _VFSH_(_vint32 n32Signal);
static _vvoid _AtExit_(_vvoid);

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Installer
// --------------------------------------------------------------------------------------------
// ***** Install fatal signal handler
_vvoid VInstallFSH(PFNVortexFatalFunc pfnUserCallback, _vvoid *pvUserData)
{
  // ..... Static variables
  static _vbool bAtExitInstalled = false;
  
  // ..... Make sure that this only installed once
  if(_bInstalled_) return;
  _bInstalled_ = true;

  // ..... Install handler
  _pfnUser_ = pfnUserCallback; _pvUserData_ = pvUserData;
  _atOldHandler_[0] = signal(SIGSEGV, _VFSH_);
  _atOldHandler_[1] = signal(SIGBUS,  _VFSH_);
  _atOldHandler_[2] = signal(SIGILL,  _VFSH_);
  _atOldHandler_[3] = signal(SIGQUIT, _VFSH_);
  _atOldHandler_[4] = signal(SIGPIPE, _VFSH_);
  _atOldHandler_[5] = signal(SIGFPE,  _VFSH_);
  
  // ..... Register a function to be called at exit
  if(!bAtExitInstalled)
  {
    bAtExitInstalled = true;
    atexit(_AtExit_);
  }
}

// ***** Uninstall fatal signal handler
_vvoid VUninstallFSH(_vvoid)
{
  // ..... Make sure that this only uninstalled once
  if(!_bInstalled_) return;
  _bInstalled_ = false;

  // ..... Uninstall handler
  _pfnUser_ = NULL; _pvUserData_ = NULL;
  signal(SIGSEGV, _atOldHandler_[0]); _atOldHandler_[0] = NULL;
  signal(SIGBUS,  _atOldHandler_[1]); _atOldHandler_[1] = NULL;
  signal(SIGILL,  _atOldHandler_[2]); _atOldHandler_[2] = NULL;
  signal(SIGQUIT, _atOldHandler_[3]); _atOldHandler_[3] = NULL;
  signal(SIGPIPE, _atOldHandler_[4]); _atOldHandler_[4] = NULL;
  signal(SIGFPE,  _atOldHandler_[5]); _atOldHandler_[5] = NULL;
}

// --------------------------------------------------------------------------------------------
// Handlers
// --------------------------------------------------------------------------------------------
// ***** The fatal signal handler
static _vvoid _VFSH_(_vint32 n32Signal)
{
  // ..... Temporary variables
  PFNVortexFatalFunc pfnTmp = _pfnUser_;
  _vvoid             *pvTmp = _pvUserData_;
  
  // ..... Restore the original handler
  VUninstallFSH();

  // ..... Print message
  if(n32Signal == SIGSEGV)      VFatalError(_F_, _L_, "Segmentation fault");
  else if(n32Signal == SIGBUS)  VFatalError(_F_, _L_, "Bus error");
  else if(n32Signal == SIGILL)  VFatalError(_F_, _L_, "Illegal instruction");
  else if(n32Signal == SIGQUIT) VFatalError(_F_, _L_, "Keyboard quit");
  else if(n32Signal == SIGPIPE) VFatalError(_F_, _L_, "Broken pipe");
  else if(n32Signal == SIGFPE)  VFatalError(_F_, _L_, "Floating point exception");
  else                          VFatalError(_F_, _L_, "Unknown signal %d", n32Signal);

  // ..... Call the user callback functions
  if(pfnTmp) pfnTmp(pvTmp);

  // ..... Exit program
  exit(-n32Signal);
}

// ***** Called at program exit to make sure that the fatal signal handler is uninstalled
static _vvoid _AtExit_(_vvoid)
{
  VUninstallFSH();
}

// ============================================================================================
// End of vfataleh.cpp
// ============================================================================================
