// ############################################################################################
//  vx11.cpp
//  Implementation of _CVortexVideoEnv class using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <vidx11/vx11.h>
#include <vidx11/vx11cur.h>

// ============================================================================================
// Local functions prototypes
// ============================================================================================
// ***** Initialization functions
_vvoid _InitX11Symbols_(_vvoid);

// ***** Coordinates-creation functions
_vvoid _CalcPanelCoords_(XRectangle *ptOut, XSegment *ptTL, XSegment *ptBR, XRectangle *ptFace,
                         _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2);
_vvoid _AdjustRectangle_(_vint32 &n32X1, _vint32 &n32Y1, _vint32 &n32X2, _vint32 &n32Y2,
                         _vint32 n32Width, _vint32 n32Height);
_vvoid _InitMsgBoxSize_(_EVortexMsgSize eSize, _vint32 &n32X1, _vint32 &n32Y1,
                        _vint32 &n32X2, _vint32 &n32Y2,
                        _vuint32 un32MaxWidth, _vuint32 un32MaxHeight);

// ============================================================================================
// Exported variables declaration
// ============================================================================================
// ***** The VortexGE symbols
_EVortexKeySymbol _aeX11VKS_[_MAX_X11_VKS_];
_EVortexBtnSymbol _aeX11VBS_[_MAX_X11_VBS_];

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The X11 display class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexVideoEnv::_CVortexVideoEnv()
{
  memset(&mt_tX11, 0, sizeof(mt_tX11));
  memset(&mt_tRI, 0, sizeof(mt_tRI));
  memset(&mt_tRB, 0, sizeof(mt_tRB));
  memset(&mt_tEB, 0, sizeof(mt_tEB));
  mt_tEB.ptEvent = &mt_tEB.tEmptyEvent;
  _InitX11Symbols_();
}
_CVortexVideoEnv::~_CVortexVideoEnv()
{
  // ..... Free the render buffer datas
  if(mt_tRB.pun32LUT)    free(mt_tRB.pun32LUT);
#ifdef _ALIGN_BUFF_MEM_
  if(mt_tRB.pn32ZBuffer) ssefree(mt_tRB.pn32ZBuffer);
  if(mt_tRB.pun8Buffer)  ssefree(mt_tRB.pun8Buffer);
#else
  if(mt_tRB.pn32ZBuffer) free(mt_tRB.pn32ZBuffer);
  if(mt_tRB.pun8Buffer)  free(mt_tRB.pun8Buffer);
#endif
  
  // ..... Free the render image datas
  if(mt_tRI.pucLUT)   free(mt_tRI.pucLUT);
  if(mt_tRI.pun16Pal) free(mt_tRI.pun16Pal);
#ifdef _ALIGN_BUFF_MEM_
  if(mt_tRI.ptImage)
  {
    // Make sure that the X frees the original address
    mt_tRI.ptImage->data = (_vchar *) ssegetrealaddress(mt_tRI.pucImage);
    XDestroyImage(mt_tRI.ptImage);
  }
  else if(mt_tRI.pucImage) 
    ssefree(mt_tRI.pucImage);
#else
  if(mt_tRI.ptImage)
    XDestroyImage(mt_tRI.ptImage);
  else if(mt_tRI.pucImage) 
    free(mt_tRI.pucImage);
#endif
  
  // ..... Free the main X11 objects
  if(mt_tX11.tMainWindow)    XUndefineCursor(mt_tX11.ptDisplay, mt_tX11.tMainWindow);
  if(mt_tX11.tDefCurArrow)   XFreeCursor(mt_tX11.ptDisplay, mt_tX11.tDefCurArrow);
  if(mt_tX11.tDefCurWait)    XFreeCursor(mt_tX11.ptDisplay, mt_tX11.tDefCurWait);
  if(mt_tX11.tMainPixmap)    XFreePixmap(mt_tX11.ptDisplay, mt_tX11.tMainPixmap);
  if(mt_tX11.tMainWindow)
  {
    XUnmapWindow(mt_tX11.ptDisplay, mt_tX11.tMainWindow);
    XDestroyWindow(mt_tX11.ptDisplay, mt_tX11.tMainWindow);
  }
  if(mt_tX11.ptDefFontBig)           XFreeFont(mt_tX11.ptDisplay, mt_tX11.ptDefFontBig);
  if(mt_tX11.ptDefFontSmall)         XFreeFont(mt_tX11.ptDisplay, mt_tX11.ptDefFontSmall);
  if(mt_tX11.tMainGC)                XFreeGC(mt_tX11.ptDisplay, mt_tX11.tMainGC);
  if(mt_tX11.tColormapData.un8Count) XFreeColormap(mt_tX11.ptDisplay, mt_tX11.tColormap);
  if(mt_tX11.ptDisplay)
  {
    XAutoRepeatOn(mt_tX11.ptDisplay);
    XCloseDisplay(mt_tX11.ptDisplay);
  }
}

// ***** Initialization functions
_vbool _CVortexVideoEnv::CreateMainWindow(_vuint32 un32Width, _vuint32 un32Height,
                                          const _vchar *pcWindowTitle, 
                                          _vbool bUseExclusiveColormap,
                                          const _vchar *pcDisplayName)
{
  // ..... Temporary variables
  XSetWindowAttributes tWA;
  XSizeHints           tSize;
  XColor               tColors[256];
  _vuint16             *pun16Tmp;
  _vint16              n16Cnt;
  _vchar               *pcDEnv;
  _vbool               bNULL;

  // ..... Connect to the X server
  pcDEnv = NULL; bNULL = false;
  if(pcDisplayName)
  {
    mt_tX11.ptDisplay = XOpenDisplay(pcDisplayName);
    if(mt_tX11.ptDisplay == NULL)
    {
      pcDEnv = getenv("DISPLAY");
      if(pcDEnv)
        VError(_F_, _L_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(\"%s\")", pcDisplayName, pcDEnv);
      else
      {
        VError(_F_, _L_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(NULL)", pcDisplayName);
        bNULL = true;
      }
    }
  }
  else
  {
    pcDEnv = getenv("DISPLAY");
    if(pcDEnv == NULL) bNULL = true;
  }
  if(pcDEnv)
  {
    mt_tX11.ptDisplay = XOpenDisplay(pcDEnv);
    if(mt_tX11.ptDisplay == NULL)
    {
      bNULL = true;
      VError(_F_, _L_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(NULL)", pcDEnv);
    }
  }
  if(bNULL) mt_tX11.ptDisplay = XOpenDisplay(NULL);
  if(mt_tX11.ptDisplay == NULL)
  {
    VSetErrorCode(VERR_XOpenDisplay);
    VError(_F_, _L_, "XOpenDisplay(NULL) failed");
    return(false);
  }
  XSynchronize(mt_tX11.ptDisplay, false);

  // ..... Save defaults
  mt_tX11.n32Screen      = DefaultScreen(mt_tX11.ptDisplay);
  mt_tX11.un32Depth      = DefaultDepth(mt_tX11.ptDisplay, mt_tX11.n32Screen);
  if((mt_tX11.un32Depth != 8) && (mt_tX11.un32Depth != 16) && (mt_tX11.un32Depth != 24))
  {
    VSetErrorCode(VERR_NotSupported);
    VError(_F_, _L_, "Current display depth (%d bpp) is not supported", mt_tX11.un32Depth);
    return(false);
  }
  mt_tX11.ptVisual       = DefaultVisual(mt_tX11.ptDisplay, mt_tX11.n32Screen);
  mt_tX11.tRootWindow    = RootWindow(mt_tX11.ptDisplay, mt_tX11.n32Screen);

  // ..... Create colormap
  _CreateColorMap_(bUseExclusiveColormap);
  
  // ..... For 8 bpp only : generate palette table & color translation LUT
  if(mt_tX11.un32Depth == 8)
  {
    // Query colors from default colormap
    for(n16Cnt = 0; n16Cnt < 256; n16Cnt++) tColors[n16Cnt].pixel = n16Cnt;
    XQueryColors(mt_tX11.ptDisplay, mt_tX11.tColormap, tColors, 256);

    // Allocate memory for the palette table
    mt_tRI.pun16Pal = (_vuint16 *) malloc(256 * 3 * sizeof(_vuint16));
    if(mt_tRI.pun16Pal == NULL)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for palette table");
      return(false);   
    }
 
    // Generate the palette table
    pun16Tmp = mt_tRI.pun16Pal;
    for(n16Cnt = 0; n16Cnt < 256; n16Cnt++)
    {
      *pun16Tmp++ = tColors[n16Cnt].red;
      *pun16Tmp++ = tColors[n16Cnt].green;
      *pun16Tmp++ = tColors[n16Cnt].blue;
    }
    
    // Allocate memory for the color translation LUT
    mt_tRI.pucLUT = (_vuchar *) malloc(32 * 32 * 32);
    if(mt_tRI.pucLUT == NULL)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for color translation LUT");
      return(false);
    }

    // Initialize the color translation LUT
    _InitColorLUT_();
  }

  // ..... Create GC
  mt_tX11.tMainGC = XCreateGC(mt_tX11.ptDisplay, mt_tX11.tRootWindow, 0, NULL);
  if(mt_tX11.tMainGC < 0)
  {
    mt_tX11.tMainGC = 0;
    VSetErrorCode(VERR_XCreateGC);
    VError(_F_, _L_, "XCreateGC failed");
    return(false);
  }
  
  // ..... Create default fonts
  mt_tX11.ptDefFontSmall = XLoadQueryFont(mt_tX11.ptDisplay, _DEFAULT_FONT_SMALL_);
  if(mt_tX11.ptDefFontSmall == NULL)
  {
    VSetErrorCode(VERR_XLoadQueryFont);
    VError(_F_, _L_, "XLoadQueryFont(%s) failed", _DEFAULT_FONT_SMALL_);
    return(false);
  }
  mt_tX11.ptDefFontBig = XLoadQueryFont(mt_tX11.ptDisplay, _DEFAULT_FONT_BIG_);
  if(mt_tX11.ptDefFontBig == NULL)
  {
    VSetErrorCode(VERR_XLoadQueryFont);
    VError(_F_, _L_, "XLoadQueryFont(%s) failed", _DEFAULT_FONT_BIG_);
    return(false);
  }
  
  // ..... Select the default small font to GC
  XSetFont(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11.ptDefFontSmall->fid);

  // ..... Generate the default style
  mt_tX11Style.ptFont              = mt_tX11.ptDefFontSmall;
  mt_tX11Style.un32WholeBorder     = CreateColor(  0,   0,   0);
  mt_tX11Style.un32LightBorder     = CreateColor(255, 255, 255);
  mt_tX11Style.un32DarkBorder      = CreateColor( 96, 96, 96);
  mt_tX11Style.un32Face            = CreateColor(164, 164, 164);
  mt_tX11Style.un32Text            = CreateColor(  0,   0,   0);
  mt_tX11Style.un32IconBorder      = CreateColor(128, 128, 128);
  mt_tX11Style.un32IconBackground  = CreateColor(  0,   0,   0);
  mt_tX11Style.un32IconInformation = CreateColor(  0,   0, 255);
  mt_tX11Style.un32IconQuestion    = CreateColor(  0, 255,   0);
  mt_tX11Style.un32IconWarning     = CreateColor(255, 255,   0);
  mt_tX11Style.un32IconError       = CreateColor(255,   0,   0);
  mt_tX11Style.n32FillStyle    = FillSolid;
  mt_tX11CS.ptFont             = mt_tX11Style.ptFont;
  mt_tX11CS.un32Foreground     = mt_tX11Style.un32LightBorder;
  mt_tX11CS.n32FillStyle       = mt_tX11Style.n32FillStyle ;
      
  // ..... Create main window
  memset(&tWA, 0, sizeof(tWA));
  tWA.border_pixel     = mt_tX11.un32WhitePixel;
  tWA.background_pixel = mt_tX11.un32BlackPixel;
  tWA.event_mask       = KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
                         PointerMotionMask | FocusChangeMask | ExposureMask | 
                         VisibilityChangeMask | StructureNotifyMask;
  tWA.colormap         = mt_tX11.tColormap;
    
  // ..... Request event using event mask
  mt_tX11.un32Width  = un32Width;
  mt_tX11.un32Height = un32Height;
  mt_tX11.tMainWindow = XCreateWindow(mt_tX11.ptDisplay, mt_tX11.tRootWindow,
                                      0, 0, un32Width, un32Height, 4, mt_tX11.un32Depth,
                                      InputOutput, mt_tX11.ptVisual,
                                      CWBorderPixel | CWBackPixel | CWEventMask | CWColormap,
                                      &tWA);

  // ..... Set main window properties
  tSize.flags = PMinSize | PMaxSize;
  tSize.min_width  = tSize.max_width  = un32Width;
  tSize.min_height = tSize.max_height = un32Height;
  XSetStandardProperties(mt_tX11.ptDisplay, mt_tX11.tMainWindow,
                         pcWindowTitle, pcWindowTitle, None, NULL, 0, &tSize);
 
  // ..... Display main window
  XMapRaised(mt_tX11.ptDisplay, mt_tX11.tMainWindow);
  
  // ..... Center main  window in screen
  XMoveWindow(mt_tX11.ptDisplay, mt_tX11.tMainWindow,
              (DisplayWidth(mt_tX11.ptDisplay, mt_tX11.n32Screen) - un32Width) / 2,
              (DisplayHeight(mt_tX11.ptDisplay, mt_tX11.n32Screen) - un32Height) / 2);

  // ..... Create default cursor
  _CreateDefaultCursor_();
  XSync(mt_tX11.ptDisplay, true);
  
  // ..... Use default "arrow" cursor
  UseDefaultArrowCursor();

  // ..... Create main pixmap
  mt_tX11.tMainPixmap = XCreatePixmap(mt_tX11.ptDisplay, mt_tX11.tMainWindow,
                                      un32Width, un32Height, mt_tX11.un32Depth);
  XSync(mt_tX11.ptDisplay, true);
  
  // ..... Clear the pixmap
  XSetForeground(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11.un32BlackPixel);
  XSetBackground(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11.un32BlackPixel);
  XFillRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 0, 0, un32Width, un32Height);
  XSetForeground(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11.un32WhitePixel);

  // ..... Set auto repeat
  AutoRepeatOn();
  
  // ..... OK !
  mt_tX11.bInitialized = true;
  mt_bActive           = true;
  mt_bVisible          = true;
  return(true);
}
_vbool _CVortexVideoEnv::CreateRenderImage(_vuint32 un32Width, _vuint32 un32Height)
{
  // ..... Temporary variable
  _vuint32 un32Size; 
  
  // ..... Check
  if(!mt_tX11.bInitialized)
  {
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Main window must be initialized before creating render image");
    return(false);
  }
  
  // ..... Calculate image size
  mt_tRI.un32Width  = ((un32Width > 0) && (un32Width <  mt_tX11.un32Width)) ?
                      un32Width  : mt_tX11.un32Width;
  mt_tRI.un32Height = ((un32Height > 0) && (un32Height <  mt_tX11.un32Height)) ?
                      un32Height : mt_tX11.un32Height;
  mt_tRI.un32Dim    = mt_tRI.un32Width * mt_tRI.un32Height;
    
  // ..... Allocate memory for image
  if(mt_tX11.un32Depth == 24)
    un32Size = (mt_tRI.un32Width * (24 / 8) + mt_tRI.un32Width) * mt_tRI.un32Height;
  else if(mt_tX11.un32Depth == 16)
    un32Size = mt_tRI.un32Width * (16 / 8) * mt_tRI.un32Height;
  else
    un32Size = mt_tRI.un32Width * (8 / 8) * mt_tRI.un32Height;
  mt_tRI.un32Size = un32Size;
#ifdef _ALIGN_BUFF_MEM_
  mt_tRI.pucImage = (_vuchar *) ssemalloc(un32Size);
#else
  mt_tRI.pucImage = (_vuchar *) malloc(un32Size);
#endif  
  if(mt_tRI.pucImage == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed allocate memory for image data");
    return(false);
  }
  memset(mt_tRI.pucImage, 0, un32Size);
  
  // ..... Create image
  if(mt_tX11.un32Depth == 24)
    mt_tRI.ptImage = XCreateImage(mt_tX11.ptDisplay, mt_tX11.ptVisual, 24, ZPixmap, 0,
                                  (char *) mt_tRI.pucImage, mt_tRI.un32Width,
                                  mt_tRI.un32Height, 8, mt_tRI.un32Width * (24 / 8) + 
                                  mt_tRI.un32Width);
  else if(mt_tX11.un32Depth == 16)
    mt_tRI.ptImage = XCreateImage(mt_tX11.ptDisplay, mt_tX11.ptVisual, 16, ZPixmap, 0,
                                  (char *) mt_tRI.pucImage, mt_tRI.un32Width,
                                  mt_tRI.un32Height, 8, mt_tRI.un32Width * (16 / 8));
  else 
    mt_tRI.ptImage = XCreateImage(mt_tX11.ptDisplay, mt_tX11.ptVisual, 8, ZPixmap, 0,
                                  (char *) mt_tRI.pucImage, mt_tRI.un32Width,
                                  mt_tRI.un32Height, 8, mt_tRI.un32Width * (8 / 8));
  if(mt_tRI.ptImage == NULL)
  {
    VSetErrorCode(VERR_XCreateImage);
    VError(_F_, _L_, "Failed to create a new XImage");
    return(false);
  }
  
  // ..... OK !
  mt_tRI.bInitialized = true;
  return(true);
}
_vbool _CVortexVideoEnv::CreateRenderBuffer(_vuint32 un32Width, _vuint32 un32Height)
{
  // ..... Check
  if(!mt_tRI.bInitialized)
  {
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Render image must be creaetd before creating render buffer");
    return(false);
  }
  
  // ..... Calculate buffer size
  mt_tRB.un32Width  = ((un32Width > 0) && (un32Width <  mt_tRI.un32Width))  ? un32Width  : mt_tRI.un32Width;
  mt_tRB.un32Height = ((un32Height > 0) && (un32Height <  mt_tRI.un32Height)) ? un32Height : mt_tRI.un32Height;

  // ..... Check buffer size
  mt_tRB.un32Size      = mt_tRB.un32Width * mt_tRB.un32Height * 4;
  mt_tRB.un32Size48    = mt_tRB.un32Size / 4 / 8;
  if((mt_tRB.un32Size48 * 4 * 8) != mt_tRB.un32Size)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Invalid render buffer size, (un32Width X un32Height) must be divisible by 8");
    return(false);
  }
  
  // ..... Allocate memory for render buffer
#ifdef _ALIGN_BUFF_MEM_
  mt_tRB.pun8Buffer = (_vuint8 *) ssemalloc(mt_tRB.un32Size);
#else
  mt_tRB.pun8Buffer = (_vuint8 *) malloc(mt_tRB.un32Size);
#endif
  if(mt_tRB.pun8Buffer == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for render buffer");
    return(false);
  }
  memset(mt_tRB.pun8Buffer, 0, mt_tRB.un32Size);
  
  // .....  Allocate memory for Z buffer
#ifdef _ALIGN_BUFF_MEM_
  mt_tRB.pn32ZBuffer = (_vint32 *) ssemalloc(mt_tRB.un32Size);
#else
  mt_tRB.pn32ZBuffer = (_vint32 *) malloc(mt_tRB.un32Size);
#endif  
  if(mt_tRB.pn32ZBuffer == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for Z buffer");
    return(false);
  }
  memset(mt_tRB.pn32ZBuffer, 0, mt_tRB.un32Size);

  // ..... Initialize some variables
  mt_tRB.un32SizeSL = mt_tRB.un32Width * 4;
  mt_tRB.fWidth2    = _vfloat(mt_tRB.un32Width) / _vfloat(2.0);
  mt_tRB.fHeight2   = _vfloat(mt_tRB.un32Height) / _vfloat(2.0);
  mt_tRB.n32ZTrans  = 0;
  
  // ..... Generate scalling LUT
  if((mt_tRB.un32Width != mt_tRI.un32Width) || ( mt_tRB.un32Height !=  mt_tRI.un32Height))
  {
    // Allocate memory
    if(mt_tX11.un32Depth == 24)
      mt_tRB.pun32LUT = (_vuint32 *) malloc(mt_tRI.un32Dim * sizeof(_vuint32));
    else
      mt_tRB.pun32LUT = (_vuint32 *) malloc(mt_tRI.un32Dim * 3 * sizeof(_vuint32));
    if(mt_tRB.pun32LUT == NULL)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for scalling LUT");
      return(false);
    }
    
    // Generate LUT
    _vfloat  fXs, fYs;
    _vuint32 un32X, un32Y, un32Xs, un32Ys;
    _vuint32 *pun32Tmp;
    
    fXs = _vfloat(mt_tRB.un32Width) / _vfloat(mt_tRI.un32Width);
    fYs = _vfloat(mt_tRB.un32Height) / _vfloat(mt_tRI.un32Height);
    
    pun32Tmp = mt_tRB.pun32LUT;
    for(un32Y = 0; un32Y < mt_tRI.un32Height; un32Y++)
    {
      un32Ys = _vuint32(un32Y * fYs);
      if(un32Ys >= mt_tRB.un32Height) un32Ys = (mt_tRB.un32Height - 1);
        
      for(un32X = 0; un32X < mt_tRI.un32Width; un32X++)
      {
        un32Xs = _vuint32(un32X * fXs);
        if(un32Xs >= mt_tRB.un32Width) un32Xs = (mt_tRB.un32Width - 1);
          
        if(mt_tX11.un32Depth == 24)
          *pun32Tmp++ = un32Ys * mt_tRB.un32Width + un32Xs;
        else
        {
          *pun32Tmp++ = (un32Ys * mt_tRB.un32Width + un32Xs) * 4 + 0;
          *pun32Tmp++ = (un32Ys * mt_tRB.un32Width + un32Xs) * 4 + 1;
          *pun32Tmp++ = (un32Ys * mt_tRB.un32Width + un32Xs) * 4 + 2;
        };
      }
    }
  }
  
  // ..... Determine the update type
  if((mt_tRB.un32Width == mt_tRI.un32Width) && (mt_tRB.un32Height == mt_tRI.un32Height))
  {
    if(mt_tX11.un32Depth == 24)
      mt_tRB.un32UT = 0;
    else if(mt_tX11.un32Depth == 16)
      mt_tRB.un32UT = 1;
    else
      mt_tRB.un32UT = 2;
  }
  else
  {
    if(mt_tX11.un32Depth == 24)
      mt_tRB.un32UT = 3;
    else if(mt_tX11.un32Depth == 16)
      mt_tRB.un32UT = 4;
    else
      mt_tRB.un32UT = 5;
  }

  // ..... OK !
  mt_bAllInitialized = true;
  return(true);
}
_vvoid _CVortexVideoEnv::_CreateColorMap_(_vbool bUseExclusiveColormap)
{
  // ..... Temporary variables
  _vuintad *punadDummy, aunadColors[216];
  _vuint32 un32C, un32R, un32G, un32B;
  _vuint32 aun32Vals[6] = {0, 13107, 26214, 39321, 52428, 65535 };
  
  // ..... Check
  if((mt_tX11.un32Depth != 8) || (!bUseExclusiveColormap))
  {
    mt_tX11.tColormap = DefaultColormap(mt_tX11.ptDisplay, mt_tX11.n32Screen);
    mt_tX11.un32BlackPixel = BlackPixel(mt_tX11.ptDisplay, mt_tX11.n32Screen);
    mt_tX11.un32WhitePixel = WhitePixel(mt_tX11.ptDisplay, mt_tX11.n32Screen);
    return;
  }
  
  // ..... Create color map
  mt_tX11.tColormap = XCreateColormap(mt_tX11.ptDisplay, mt_tX11.tRootWindow, mt_tX11.ptVisual, AllocNone);
  
  // ..... Allocate memory
  mt_tX11.tColormapData.ptColors = (XColor *) malloc(sizeof(XColor) * 216);
  if(mt_tX11.tColormapData.ptColors == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    mt_tX11.tColormap = DefaultColormap(mt_tX11.ptDisplay, mt_tX11.n32Screen);
    VError(_F_, _L_, "Cannot allocate memory for colormap colors");
    return;
  }
  
  // ..... Allocate color cells
  punadDummy = NULL;
  if(!XAllocColorCells(mt_tX11.ptDisplay, mt_tX11.tColormap, 0, punadDummy, 0,
                       aunadColors, 216))
  {
    free(mt_tX11.tColormapData.ptColors);
    mt_tX11.tColormap = DefaultColormap(mt_tX11.ptDisplay, mt_tX11.n32Screen);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Cannot allocate color cells");
    return;
  }
  
  // ..... Create color data
  for(un32C = 0; un32C < 216; un32C++)
  {
    mt_tX11.tColormapData.ptColors[un32C].pixel = aunadColors[un32C];
    mt_tX11.tColormapData.ptColors[un32C].flags = DoRed | DoGreen | DoBlue;
  }
  un32C = 0 ;
  for(un32R = 0; un32R < 6; un32R++)
    for(un32G = 0; un32G < 6; un32G++)
      for(un32B = 0; un32B < 6; un32B++)
      {
        mt_tX11.tColormapData.ptColors[un32C].red   = aun32Vals[un32R];
        mt_tX11.tColormapData.ptColors[un32C].green = aun32Vals[un32G];
        mt_tX11.tColormapData.ptColors[un32C].blue  = aun32Vals[un32B];
        un32C++;
      }
      
  // ..... Store colors
  XStoreColors(mt_tX11.ptDisplay, mt_tX11.tColormap, mt_tX11.tColormapData.ptColors, 216);
  mt_tX11.un32BlackPixel = 0;
  mt_tX11.un32WhitePixel = 215;
  mt_tX11.tColormapData.un8Count = 216;
}  
_vvoid _CVortexVideoEnv::_InitColorLUT_(_vvoid)
{
  // ..... Temporary variables
  _vuint16 un8R, un8G, un8B, un8C, un8CSel;
  _vuint16 un16R1, un16G1, un16B1;
  _vuint16 un16R2, un16G2, un16B2;
  _vuint16 *pun16Tmp;
  _vuint32 un32DR, un32DG, un32DB, un32D, un32DMin;
  
  // ..... Generate LUT
  for(un8R = 0; un8R <= 31; un8R++)
    for(un8G = 0; un8G <= 31; un8G++)
      for(un8B = 0; un8B <= 31; un8B++)
      {
        // Scale color
        un16R1 = _vuint16(un8R) * _vuint16(2114);
        un16G1 = _vuint16(un8G) * _vuint16(2114);
        un16B1 = _vuint16(un8B) * _vuint16(2114);
        
        // Search for closest in the palette
        un8CSel  = 0;
        un32DMin = 10000;
        pun16Tmp = mt_tRI.pun16Pal;
        for(un8C = 0; un8C <= 255; un8C++)
        {
          un16R2 = *pun16Tmp++;
          un16G2 = *pun16Tmp++;
          un16B2 = *pun16Tmp++;

          un32DR = (un16R2 - un16R1) >> 3;
          un32DG = (un16G2 - un16G1) >> 3;
          un32DB = (un16B2 - un16B1) >> 3;
        
          un32D  = (_vuint32) sqrt((un32DR * un32DR) + (un32DG * un32DG) + (un32DB * un32DB));
 
          if(un32D < un32DMin)
          {
            un32DMin = un32D;
            un8CSel  = un8C;
          }
        }
        
        // Save the result in LUT
        mt_tRI.pucLUT[(_vuint32(un8R) << 10) | (_vuint32(un8G) << 5) | _vuint32(un8B)] = un8CSel;
      }
}
_vvoid _CVortexVideoEnv::_CreateDefaultCursor_(_vvoid)
{
  // ..... Temporary variables
  XColor tDummy, tFore, tBack;
  Pixmap tTmp, tCursor, tMask;
  _vuint16 an16Arrow[]     = _VX11CUR_ARROW_;
  _vuint16 an16ArrowMask[] = _VX11CUR_ARROW_MASK_;
  _vuint16 an16Wait[]      = _VX11CUR_WAIT_;
  _vuint16 an16WaitMask[]  = _VX11CUR_WAIT_MASK_;
  
  // ..... Create default "arrow" cursor
  XLookupColor(mt_tX11.ptDisplay, mt_tX11.tColormap, "white", &tDummy, &tFore);
  XLookupColor(mt_tX11.ptDisplay, mt_tX11.tColormap, "black", &tDummy, &tBack);
  tTmp    = XCreatePixmap(mt_tX11.ptDisplay, mt_tX11.tRootWindow, 1, 1, 1);
  tCursor = XCreatePixmapFromBitmapData(mt_tX11.ptDisplay, tTmp, (_vchar *) an16Arrow,
                                        _VX11CUR_ARROW_X_SIZE_, _VX11CUR_ARROW_Y_SIZE_, 1, 0, 1);
  tMask   = XCreatePixmapFromBitmapData(mt_tX11.ptDisplay, tTmp, (_vchar *) an16ArrowMask,
                                        _VX11CUR_ARROW_X_SIZE_, _VX11CUR_ARROW_Y_SIZE_, 1, 0, 1);
  mt_tX11.tDefCurArrow = XCreatePixmapCursor(mt_tX11.ptDisplay, tCursor, tMask, &tFore, &tBack,
                                             _VX11CUR_ARROW_X_HP_, _VX11CUR_ARROW_Y_HP_);
  XFreePixmap(mt_tX11.ptDisplay, tTmp);
  XFreePixmap(mt_tX11.ptDisplay, tCursor);
  XFreePixmap(mt_tX11.ptDisplay, tMask);

  // ..... Create default "wait" cursor
  XLookupColor(mt_tX11.ptDisplay, mt_tX11.tColormap, "white", &tDummy, &tFore);
  XLookupColor(mt_tX11.ptDisplay, mt_tX11.tColormap, "black", &tDummy, &tBack);
  tTmp    = XCreatePixmap(mt_tX11.ptDisplay, mt_tX11.tRootWindow, 1, 1, 1);
  tCursor = XCreatePixmapFromBitmapData(mt_tX11.ptDisplay, tTmp, (_vchar *) an16Wait,
                                        _VX11CUR_WAIT_X_SIZE_, _VX11CUR_WAIT_Y_SIZE_, 1, 0, 1);
  tMask   = XCreatePixmapFromBitmapData(mt_tX11.ptDisplay, tTmp, (_vchar *) an16WaitMask,
                                        _VX11CUR_WAIT_X_SIZE_, _VX11CUR_WAIT_Y_SIZE_, 1, 0, 1);
  mt_tX11.tDefCurWait = XCreatePixmapCursor(mt_tX11.ptDisplay, tCursor, tMask, &tFore, &tBack,
                                           _VX11CUR_WAIT_X_HP_, _VX11CUR_WAIT_Y_HP_);
  XFreePixmap(mt_tX11.ptDisplay, tTmp);
  XFreePixmap(mt_tX11.ptDisplay, tCursor);
  XFreePixmap(mt_tX11.ptDisplay, tMask);
}

// ***** Creation functions
_vbool _CVortexVideoEnv::CreateButton(_CVortexButton &csVButton, 
                                      _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2,
                                      const _vchar *pcText, _vbool bInsideOutline,
                                      _vbool bThickOutline, _vbool bDuplicateString)
{
  // ..... Calculate panels's coordinates
  if(bThickOutline)
  {
    n32X1--; n32Y1--;
    n32X2++; n32Y2++;
    csVButton.mt_bThickOutline = true;
  }
  else
    csVButton.mt_bThickOutline = false;
  _CalcPanelCoords_(&csVButton.mt_tOut, csVButton.mt_atSegTL,
                    csVButton.mt_atSegBR, &csVButton.mt_tFace,
                    n32X1, n32Y1, n32X2, n32Y2);
  if(bInsideOutline)
  {
    csVButton.mt_tIn.x = csVButton.mt_tFace.x++;
    csVButton.mt_tIn.y = csVButton.mt_tFace.y++;
    csVButton.mt_tFace.width--;
    csVButton.mt_tFace.height--;
    csVButton.mt_tIn.width  = csVButton.mt_tFace.width--;
    csVButton.mt_tIn.height = csVButton.mt_tFace.height--;
    csVButton.mt_bInsideOutline = true;
  }
  else
    csVButton.mt_bInsideOutline = false;
    
  csVButton.mt_tOutThick.x = csVButton.mt_tOut.x - 1;
  csVButton.mt_tOutThick.y = csVButton.mt_tOut.y - 1;
  csVButton.mt_tOutThick.width  = csVButton.mt_tOut.width + 2;
  csVButton.mt_tOutThick.height = csVButton.mt_tOut.height + 2;
  
  // ..... Calculate button's text properties
  csVButton.mt_un32TextLen = strlen(pcText);
  csVButton.mt_n32Xt = n32X1 + (n32X2 - n32X1 - XTextWidth(mt_tX11Style.ptFont, pcText, csVButton.mt_un32TextLen)) / 2;
  csVButton.mt_n32Yt = n32Y1 + (n32Y2 - n32Y1 - mt_tX11Style.ptFont->ascent - mt_tX11Style.ptFont->descent) / 2 + mt_tX11.ptDefFontSmall->ascent;

  // ..... Copy text
  if((csVButton.mt_bDuplicateString) && (csVButton.mt_pcText != NULL)) free(csVButton.mt_pcText);
  csVButton.mt_bDuplicateString= false;  
  if(bDuplicateString)
  {
    csVButton.mt_pcText = strdup(pcText);
    if(csVButton.mt_pcText == NULL)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for new button");
      return(false);
    }
    csVButton.mt_bDuplicateString = true;  
  }
  else
    csVButton.mt_pcText = (_vchar *) pcText;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexVideoEnv::CreatePanelBar(_CVortexPanelBar &csVPanelBar, 
                                        _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2,
                                        const _vchar *pcCaption, _vbool bDuplicateString)
{
  // ..... Temporary variables
  _vint32 n32FH, n32FY, n32CW;

  // ..... Calculate panels's coordinates
  n32FH = mt_tX11Style.ptFont->ascent + mt_tX11Style.ptFont->descent;
  n32FY = n32Y1 + n32FH + 10;
  _CalcPanelCoords_(csVPanelBar.mt_atOut, csVPanelBar.mt_atSegTL, 
                    csVPanelBar.mt_atSegBR, csVPanelBar.mt_atFace, 
                    n32X1, n32Y1, n32X2, n32FY);
  _CalcPanelCoords_(&csVPanelBar.mt_atOut[1], &csVPanelBar.mt_atSegTL[4], 
                    &csVPanelBar.mt_atSegBR[4], &csVPanelBar.mt_atFace[1],
                    n32X1, n32FY, n32X2, n32Y2);
  csVPanelBar.mt_atSegLL[0].x1 = n32X1 + 8; csVPanelBar.mt_atSegLL[0].y1 = n32Y1 + 8;
  csVPanelBar.mt_atSegLL[0].x2 = n32X2 - 8; csVPanelBar.mt_atSegLL[0].y2 = n32Y1 + 8;
  csVPanelBar.mt_atSegLL[1].x1 = n32X1 + 8; csVPanelBar.mt_atSegLL[1].y1 = n32FY - 8;
  csVPanelBar.mt_atSegLL[1].x2 = n32X2 - 8; csVPanelBar.mt_atSegLL[1].y2 = n32FY - 8;
  csVPanelBar.mt_atSegLL[2].x1 = n32X1 + 8; csVPanelBar.mt_atSegLL[2].y1 = n32Y1 + (n32FY - n32Y1) / 2;
  csVPanelBar.mt_atSegLL[2].x2 = n32X2 - 8; csVPanelBar.mt_atSegLL[2].y2 = n32Y1 + (n32FY - n32Y1) / 2;
  csVPanelBar.mt_atSegLL[3].x1 = csVPanelBar.mt_atSegLL[0].x1; csVPanelBar.mt_atSegLL[3].y1 = csVPanelBar.mt_atSegLL[0].y1 + 1;
  csVPanelBar.mt_atSegLL[3].x2 = csVPanelBar.mt_atSegLL[0].x2; csVPanelBar.mt_atSegLL[3].y2 = csVPanelBar.mt_atSegLL[0].y2 + 1;
  csVPanelBar.mt_atSegLL[4].x1 = csVPanelBar.mt_atSegLL[1].x1; csVPanelBar.mt_atSegLL[4].y1 = csVPanelBar.mt_atSegLL[1].y1 + 1;
  csVPanelBar.mt_atSegLL[4].x2 = csVPanelBar.mt_atSegLL[1].x2; csVPanelBar.mt_atSegLL[4].y2 = csVPanelBar.mt_atSegLL[1].y2 + 1;
  csVPanelBar.mt_atSegLL[5].x1 = csVPanelBar.mt_atSegLL[2].x1; csVPanelBar.mt_atSegLL[5].y1 = csVPanelBar.mt_atSegLL[2].y1 + 1;
  csVPanelBar.mt_atSegLL[5].x2 = csVPanelBar.mt_atSegLL[2].x2; csVPanelBar.mt_atSegLL[5].y2 = csVPanelBar.mt_atSegLL[2].y2 + 1;

  // ..... Calculate panel-bar's client-start-position
  csVPanelBar.mt_n32Xt = n32X1 + 15;
  csVPanelBar.mt_n32Yt = n32FY + 15;

  // ..... Calculate panel-bar's caption properties
  csVPanelBar.mt_un32CaptionLen = strlen(pcCaption);
  n32CW = XTextWidth(mt_tX11Style.ptFont, pcCaption, csVPanelBar.mt_un32CaptionLen);
  csVPanelBar.mt_n32XC  = n32X1 + (n32X2 - n32X1 - n32CW) / 2 - 6;
  csVPanelBar.mt_n32XCW = n32CW + 12;
  csVPanelBar.mt_n32YC  = n32Y1 + mt_tX11Style.ptFont->ascent + 5;

  // ..... Copy text
  if((csVPanelBar.mt_bDuplicateString) && (csVPanelBar.mt_pcCaption != NULL)) free(csVPanelBar.mt_pcCaption);
  csVPanelBar.mt_bDuplicateString= false;  
  if(bDuplicateString)
  {
    csVPanelBar.mt_pcCaption = strdup(pcCaption);
    if(csVPanelBar.mt_pcCaption == NULL)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for new panel-bar");
      return(false);
    }
    csVPanelBar.mt_bDuplicateString = true;  
  }
  else
    csVPanelBar.mt_pcCaption= (_vchar *) pcCaption;
  
  // ..... OK !
  return(true);
}

// ***** Simple dialog functions
_EVortexMsgRetVal _CVortexVideoEnv::MessageBox(const _vchar *pcCaption, const _vchar *pcText,
                                               _EVortexMsgIcon eIcon, _EVortexMsgButton eButton,
                                               _EVortexMsgSize eSize)
{
  // ..... Temporary variables
  _EVortexMsgRetVal eTmp;
  
  // ..... Clear events
  usleep(100000);
  ClearEvent();
  
  // ..... Show message box
  if(eButton == EBtnOKCancel)
    eTmp = _OCYNARBox_(pcCaption, pcText, eIcon, eSize, 0);
  else if(eButton == EBtnYesNo)
    eTmp = _OCYNARBox_(pcCaption, pcText, eIcon, eSize, 1);
  else if(eButton == EBtnYesNoCancel)
    eTmp = _YNCARIBox_(pcCaption, pcText, eIcon, eSize, 0);
  else if(eButton == EBtnAbortRetry)
    eTmp = _OCYNARBox_(pcCaption, pcText, eIcon, eSize, 2);
  else if(eButton == EBtnAbortRetryIgnore)
    eTmp = _YNCARIBox_(pcCaption, pcText, eIcon, eSize, 1);
  else
  {
    _OBox_(pcCaption, pcText, eIcon, eSize);
    eTmp = EOK;
  }

  // ..... Clear events
  usleep(100000);
  ClearEvent();
  
  // ..... Return values
  return(eTmp);
}
_vvoid _CVortexVideoEnv::_OBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize)
{
  // ..... Temporary variables
  _CVortexPanelBar csPBar;
  _CVortexButton   csBtnOK;
  _vint32          n32X1, n32Y1, n32X2, n32Y2, n32FH, n32YC;
  KeySym           tKey;

  // ..... Initialize dialog
  _InitMsgBoxSize_(eSize, n32X1, n32Y1, n32X2, n32Y2, mt_tRI.un32Width, mt_tRI.un32Height);
  CreatePanelBar(csPBar, n32X1, n32Y1, n32X2, n32Y2, pcCaption, false);
  n32FH = mt_tX11Style.ptFont->ascent + mt_tX11Style.ptFont->descent;
  n32YC = n32Y2 - n32FH - 15;
  CreateButton(csBtnOK, n32X1 + 15, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "OK", false, true, false);
 
  // ..... Draw the dialog
  DrawPanelBar(csPBar);
  DrawMsgIcon(csPBar.mt_n32Xt, csPBar.mt_n32Yt, eIcon);
  SetFont(mt_tX11Style.ptFont);
  DrawTextML(csPBar.mt_n32Xt + 32 + 10, csPBar.mt_n32Yt + mt_tX11Style.ptFont->ascent, pcText);
  DrawButton(csBtnOK);
  UpdateMainWindow();              
  
  // ..... Loop
  while(1)
  {
    BlockedDispatchEvent();
    if(CheckEvent())
    {
      // Expose event
      if((mt_tEB.ptEvent->type == Expose) && (mt_tEB.ptEvent->xexpose.count == 0))
        UpdateMainWindow();              
      // Keyboard event
      else if(mt_tEB.ptEvent->type == KeyRelease)
      {
        tKey = _TranslateKeySymbol_(mt_tEB.ptEvent);
        if(tKey == XK_space) return;
        if(tKey == XK_Return) return;
        if(tKey == XK_Escape) return;
      }
      // Mouse event
      else if(mt_tEB.ptEvent->type == MotionNotify)
      {
        DrawButton(csBtnOK, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonPress)
      {
        DrawButton(csBtnOK, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonRelease)
      {
        DrawButton(csBtnOK, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        UpdateMainWindow();
        if(csBtnOK.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1)) return;
      }
    }
  }
}
_EVortexMsgRetVal _CVortexVideoEnv::_OCYNARBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize, _vuint8 un8Type)
{
  // ..... Temporary variables
  _CVortexPanelBar csPBar;
  _CVortexButton   csBtnOYA, csBtnCNR;
  _vint32          n32X1, n32Y1, n32X2, n32Y2, n32FH, n32XC, n32YC;
  KeySym           tKey;

  // ..... Initialize dialog
  _InitMsgBoxSize_(eSize, n32X1, n32Y1, n32X2, n32Y2, mt_tRI.un32Width, mt_tRI.un32Height);
  CreatePanelBar(csPBar, n32X1, n32Y1, n32X2, n32Y2, pcCaption, false);
  n32FH = mt_tX11Style.ptFont->ascent + mt_tX11Style.ptFont->descent;
  n32YC = n32Y2 - n32FH - 15;
  n32XC = n32X1 + (n32X2 - n32X1) / 2;
  if(un8Type == 0)
  {
    CreateButton(csBtnOYA, n32X1 + 15, n32YC - n32FH, n32XC - 7, n32YC + n32FH, "OK", false, true, false);
    CreateButton(csBtnCNR, n32XC + 7, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "Cancel", true, false, false);
  }
  else if(un8Type == 1)
  {
    CreateButton(csBtnOYA, n32X1 + 15, n32YC - n32FH, n32XC - 7, n32YC + n32FH, "Yes", false, true, false);
    CreateButton(csBtnCNR, n32XC + 7, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "No", true, false, false);
  }
  else
  {
    CreateButton(csBtnOYA, n32X1 + 15, n32YC - n32FH, n32XC - 7, n32YC + n32FH, "Abort", true, false, false);
    CreateButton(csBtnCNR, n32XC + 7, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "Retry", false, true, false);
  }
 
  // ..... Draw the dialog
  DrawPanelBar(csPBar);
  DrawMsgIcon(csPBar.mt_n32Xt, csPBar.mt_n32Yt, eIcon);
  SetFont(mt_tX11Style.ptFont);
  DrawTextML(csPBar.mt_n32Xt + 32 + 10, csPBar.mt_n32Yt + mt_tX11Style.ptFont->ascent, pcText);
  DrawButton(csBtnOYA);
  DrawButton(csBtnCNR);
  UpdateMainWindow();              
  
  // ..... Loop
  while(1)
  {
    BlockedDispatchEvent();
    if(CheckEvent())
    {
      // Expose event
      if((mt_tEB.ptEvent->type == Expose) && (mt_tEB.ptEvent->xexpose.count == 0))
        UpdateMainWindow();              
      // Keyboard event
      else if(mt_tEB.ptEvent->type == KeyRelease)
      {
        tKey = _TranslateKeySymbol_(mt_tEB.ptEvent);
        if(tKey == XK_Return)
        {
          if(un8Type == 0)
            return(EOK);
          else if(un8Type == 1)
            return(EYes);
          else
            return(ERetry);
        }
        if(tKey == XK_Escape)
        {
          if(un8Type == 0)
            return(ECancel);
          else if(un8Type == 1)
            return(ENo);
          else
            return(EAbort);
        }
      }
      // Mouse event
      else if(mt_tEB.ptEvent->type == MotionNotify)
      {
        DrawButton(csBtnOYA, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        DrawButton(csBtnCNR, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonPress)
      {
        DrawButton(csBtnOYA, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        DrawButton(csBtnCNR, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonRelease)
      {
        DrawButton(csBtnOYA, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        DrawButton(csBtnCNR, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        UpdateMainWindow();
        if(csBtnOYA.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1))
        {
          if(un8Type == 0)
            return(EOK);
          else if(un8Type == 1)
            return(EYes);
          else
            return(EAbort);
        }
        if(csBtnCNR.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1))
        {
          if(un8Type == 0)
            return(ECancel);
          else if(un8Type == 1)
            return(ENo);
          else
            return(ERetry);
        }
      }
    }
  }
}
_EVortexMsgRetVal _CVortexVideoEnv::_YNCARIBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize, _vuint8 un8Type)
{
  // ..... Temporary variables
  _CVortexPanelBar csPBar;
  _CVortexButton   csBtnYA, csBtnNR, csBtnCI;
  _vint32          n32X1, n32Y1, n32X2, n32Y2, n32FH, n32XC1, n32XC2, n32YC;
  KeySym           tKey;

  // ..... Initialize dialog
  _InitMsgBoxSize_(eSize, n32X1, n32Y1, n32X2, n32Y2, mt_tRI.un32Width, mt_tRI.un32Height);
  CreatePanelBar(csPBar, n32X1, n32Y1, n32X2, n32Y2, pcCaption, false);
  n32FH = mt_tX11Style.ptFont->ascent + mt_tX11Style.ptFont->descent;
  n32YC = n32Y2 - n32FH - 15;
  n32XC1 = n32X1 + (n32X2 - n32X1) / 3;
  n32XC2 = n32X1 + (n32X2 - n32X1) / 3 * 2;
  if(un8Type == 0)
  {
    CreateButton(csBtnYA, n32X1 + 15, n32YC - n32FH, n32XC1 - 7, n32YC + n32FH, "Yes", false, true, false);
    CreateButton(csBtnNR, n32XC1 + 7, n32YC - n32FH, n32XC2 - 7, n32YC + n32FH, "No", false, false, false);
    CreateButton(csBtnCI, n32XC2 + 7, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "Cancel", true, false, false);
  }
  else
  {
    CreateButton(csBtnYA, n32X1 + 15, n32YC - n32FH, n32XC1 - 7, n32YC + n32FH, "Abort", true, false, false);
    CreateButton(csBtnNR, n32XC1 + 7, n32YC - n32FH, n32XC2 - 7, n32YC + n32FH, "Retry", false, true, false);
    CreateButton(csBtnCI, n32XC2 + 7, n32YC - n32FH, n32X2 - 15, n32YC + n32FH, "Ignore", false, false, false);
  }
 
  // ..... Draw the dialog
  DrawPanelBar(csPBar);
  DrawMsgIcon(csPBar.mt_n32Xt, csPBar.mt_n32Yt, eIcon);
  SetFont(mt_tX11Style.ptFont);
  DrawTextML(csPBar.mt_n32Xt + 32 + 10, csPBar.mt_n32Yt + mt_tX11Style.ptFont->ascent, pcText);
  DrawButton(csBtnYA);
  DrawButton(csBtnNR);
  DrawButton(csBtnCI);
  UpdateMainWindow();              
  
  // ..... Loop
  while(1)
  {
    BlockedDispatchEvent();
    if(CheckEvent())
    {
      // Expose event
      if((mt_tEB.ptEvent->type == Expose) && (mt_tEB.ptEvent->xexpose.count == 0))
        UpdateMainWindow();              
      // Keyboard event
      else if(mt_tEB.ptEvent->type == KeyRelease)
      {
        tKey = _TranslateKeySymbol_(mt_tEB.ptEvent);
        if(tKey == XK_Return)
        {
          if(un8Type == 0)
            return(EYes);
          else
            return(ERetry);
        }
        if(tKey == XK_Escape)
        {
          if(un8Type == 0)
            return(ECancel);
          else
            return(EAbort);
        }
      }
      // Mouse event
      else if(mt_tEB.ptEvent->type == MotionNotify)
      {
        DrawButton(csBtnYA, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        DrawButton(csBtnNR, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        DrawButton(csBtnCI, mt_tEB.ptEvent->xmotion.x, mt_tEB.ptEvent->xmotion.y, mt_tEB.ptEvent->xmotion.state & Button1Mask);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonPress)
      {
        DrawButton(csBtnYA, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        DrawButton(csBtnNR, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        DrawButton(csBtnCI, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, mt_tEB.ptEvent->xbutton.button == Button1);
        UpdateMainWindow();
      }
      else if(mt_tEB.ptEvent->type == ButtonRelease)
      {
        DrawButton(csBtnYA, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        DrawButton(csBtnNR, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        DrawButton(csBtnCI, mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y, false);
        UpdateMainWindow();
        if(csBtnYA.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1))
        {
          if(un8Type == 0)
            return(EYes);
          else
            return(EAbort);
        }
        if(csBtnNR.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1))
        {
          if(un8Type == 0)
            return(ENo);
          else
            return(ERetry);
        }
        if(csBtnCI.MouseOver(mt_tEB.ptEvent->xbutton.x, mt_tEB.ptEvent->xbutton.y) && (mt_tEB.ptEvent->xbutton.button == Button1))
        {
          if(un8Type == 0)
            return(ECancel);
          else
            return(EIgnore);
        }
      }
    }
  }
}

// ============================================================================================
// Local functions implementation
// ============================================================================================
// ***** Initialization functions
_vvoid _InitX11Symbols_(_vvoid)
{
  // ..... Static variables
  _vbool bInitialized = false;
  
  // ..... Temporary variables
  _vuint16 un16Cnt;
  
  // ..... Check
  if(bInitialized) return;
  bInitialized = true;
  
  // ..... Initialize key symbol
  for(un16Cnt = 0; un16Cnt < _MAX_X11_VKS_; un16Cnt++) _aeX11VKS_[un16Cnt] = VKS_UNKNOWN;
  _aeX11VKS_[XK_BackSpace    & 0x00FF] = VKS_BS;
  _aeX11VKS_[XK_Tab          & 0x00FF] = VKS_TAB;
  _aeX11VKS_[XK_Return       & 0x00FF] = VKS_ENTER;
  _aeX11VKS_[XK_Pause        & 0x00FF] = VKS_PAUSE;
  _aeX11VKS_[XK_Sys_Req      & 0x00FF] = VKS_SYS_REQ;
  _aeX11VKS_[XK_Escape       & 0x00FF] = VKS_ESC;
  _aeX11VKS_[XK_Delete       & 0x00FF] = VKS_DEL;
  _aeX11VKS_[XK_Print        & 0x00FF] = VKS_PRINT;
  _aeX11VKS_[XK_KP_Decimal   & 0x00FF] = VKS_KP_PERIOD;
  _aeX11VKS_[XK_KP_0         & 0x00FF] = VKS_KP_0;
  _aeX11VKS_[XK_KP_1         & 0x00FF] = VKS_KP_1;
  _aeX11VKS_[XK_KP_2         & 0x00FF] = VKS_KP_2;
  _aeX11VKS_[XK_KP_3         & 0x00FF] = VKS_KP_3;
  _aeX11VKS_[XK_KP_4         & 0x00FF] = VKS_KP_4;
  _aeX11VKS_[XK_KP_5         & 0x00FF] = VKS_KP_5;
  _aeX11VKS_[XK_KP_6         & 0x00FF] = VKS_KP_6;
  _aeX11VKS_[XK_KP_7         & 0x00FF] = VKS_KP_7;
  _aeX11VKS_[XK_KP_8         & 0x00FF] = VKS_KP_8;
  _aeX11VKS_[XK_KP_9         & 0x00FF] = VKS_KP_9;
  _aeX11VKS_[XK_KP_Insert    & 0x00FF] = VKS_KP_0;
  _aeX11VKS_[XK_KP_End       & 0x00FF] = VKS_KP_1;
  _aeX11VKS_[XK_KP_Down      & 0x00FF] = VKS_KP_2;
  _aeX11VKS_[XK_KP_Page_Down & 0x00FF] = VKS_KP_3;
  _aeX11VKS_[XK_KP_Left      & 0x00FF] = VKS_KP_4;
  _aeX11VKS_[XK_KP_Begin     & 0x00FF] = VKS_KP_5;
  _aeX11VKS_[XK_KP_Right     & 0x00FF] = VKS_KP_6;
  _aeX11VKS_[XK_KP_Home      & 0x00FF] = VKS_KP_7;
  _aeX11VKS_[XK_KP_Up        & 0x00FF] = VKS_KP_8;
  _aeX11VKS_[XK_KP_Page_Up   & 0x00FF] = VKS_KP_9;
  _aeX11VKS_[XK_Delete       & 0x00FF] = VKS_KP_PERIOD;
  _aeX11VKS_[XK_KP_Divide    & 0x00FF] = VKS_KP_DIV;
  _aeX11VKS_[XK_KP_Multiply  & 0x00FF] = VKS_KP_MUL;
  _aeX11VKS_[XK_KP_Subtract  & 0x00FF] = VKS_KP_SUB;
  _aeX11VKS_[XK_KP_Add       & 0x00FF] = VKS_KP_ADD;
  _aeX11VKS_[XK_KP_Equal     & 0x00FF] = VKS_KP_EQUAL;
  _aeX11VKS_[XK_KP_Enter     & 0x00FF] = VKS_KP_ENTER;
  _aeX11VKS_[XK_Up           & 0x00FF] = VKS_UP;
  _aeX11VKS_[XK_Down         & 0x00FF] = VKS_DOWN;
  _aeX11VKS_[XK_Left         & 0x00FF] = VKS_LEFT;
  _aeX11VKS_[XK_Right        & 0x00FF] = VKS_RIGHT;
  _aeX11VKS_[XK_Insert       & 0x00FF] = VKS_INS;
  _aeX11VKS_[XK_Home         & 0x00FF] = VKS_HOME;
  _aeX11VKS_[XK_End          & 0x00FF] = VKS_END;
  _aeX11VKS_[XK_Page_Up      & 0x00FF] = VKS_PGUP;
  _aeX11VKS_[XK_Page_Down    & 0x00FF] = VKS_PGDN;
  _aeX11VKS_[XK_F1           & 0x00FF] = VKS_F1;
  _aeX11VKS_[XK_F2           & 0x00FF] = VKS_F2;
  _aeX11VKS_[XK_F3           & 0x00FF] = VKS_F3;
  _aeX11VKS_[XK_F4           & 0x00FF] = VKS_F4;
  _aeX11VKS_[XK_F5           & 0x00FF] = VKS_F5;
  _aeX11VKS_[XK_F6           & 0x00FF] = VKS_F6;
  _aeX11VKS_[XK_F7           & 0x00FF] = VKS_F7;
  _aeX11VKS_[XK_F8           & 0x00FF] = VKS_F8;
  _aeX11VKS_[XK_F9           & 0x00FF] = VKS_F9;
  _aeX11VKS_[XK_F10          & 0x00FF] = VKS_F10;
  _aeX11VKS_[XK_F11          & 0x00FF] = VKS_F11;
  _aeX11VKS_[XK_F12          & 0x00FF] = VKS_F12;
  _aeX11VKS_[XK_F13          & 0x00FF] = VKS_F13;
  _aeX11VKS_[XK_F14          & 0x00FF] = VKS_F14;
  _aeX11VKS_[XK_F15          & 0x00FF] = VKS_F15;
  _aeX11VKS_[XK_Caps_Lock    & 0x00FF] = VKS_CAPS_LOCK;
  _aeX11VKS_[XK_Num_Lock     & 0x00FF] = VKS_NUM_LOCK_;
  _aeX11VKS_[XK_Scroll_Lock  & 0x00FF] = VKS_SCROLL_LOCK;
  _aeX11VKS_[XK_Shift_L      & 0x00FF] = VKS_LSHIFT;
  _aeX11VKS_[XK_Shift_R      & 0x00FF] = VKS_RSHIFT;
  _aeX11VKS_[XK_Control_L    & 0x00FF] = VKS_LCTRL;
  _aeX11VKS_[XK_Control_R    & 0x00FF] = VKS_RCTRL;
  _aeX11VKS_[XK_Alt_R        & 0x00FF] = VKS_LALT;
  _aeX11VKS_[XK_Alt_L        & 0x00FF] = VKS_RALT;
  _aeX11VKS_[XK_Menu         & 0x00FF] = VKS_MENU;

  // ..... Initialize button symbol
  for(un16Cnt = 0; un16Cnt < _MAX_X11_VBS_; un16Cnt++) _aeX11VBS_[un16Cnt] = VBS_UNKNOWN;
  _aeX11VBS_[Button1] = VBS_MB1;
  _aeX11VBS_[Button2] = VBS_MB2;
  _aeX11VBS_[Button3] = VBS_MB3;
  _aeX11VBS_[Button4] = VBS_MB4;
  _aeX11VBS_[Button5] = VBS_MB5;
}

// *****  Coordinates-creation functions
_vvoid _CalcPanelCoords_(XRectangle *ptOut, XSegment *ptTL, XSegment *ptBR, XRectangle *ptFace,
                         _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2)
{
  ptOut->x = n32X1;
  ptOut->y = n32Y1;
  ptOut->width  = n32X2 - n32X1;
  ptOut->height = n32Y2 - n32Y1;
  n32X1++; n32Y1++; n32X2--; n32Y2--;
  ptTL[0].x1 = n32X1;     ptTL[0].y1 = n32Y1;
  ptTL[0].x2 = n32X2 - 1; ptTL[0].y2 = n32Y1;
  ptTL[1].x1 = n32X1;     ptTL[1].y1 = n32Y1 + 1;
  ptTL[1].x2 = n32X2 - 1; ptTL[1].y2 = n32Y1 + 1;
  ptTL[2].x1 = n32X1;     ptTL[2].y1 = n32Y1 + 2;
  ptTL[2].x2 = n32X1;     ptTL[2].y2 = n32Y2;
  ptTL[3].x1 = n32X1 + 1; ptTL[3].y1 = n32Y1 + 2;
  ptTL[3].x2 = n32X1 + 1; ptTL[3].y2 = n32Y2 - 1;
  ptBR[0].x1 = n32X2;     ptBR[0].y1 = n32Y1;
  ptBR[0].x2 = n32X2;     ptBR[0].y2 = n32Y2;
  ptBR[1].x1 = n32X2 - 1; ptBR[1].y1 = n32Y1 + 1;
  ptBR[1].x2 = n32X2 - 1; ptBR[1].y2 = n32Y2;
  ptBR[2].x1 = n32X1 + 1; ptBR[2].y1 = n32Y2;
  ptBR[2].x2 = n32X2 - 2; ptBR[2].y2 = n32Y2;
  ptBR[3].x1 = n32X1 + 2; ptBR[3].y1 = n32Y2 - 1;
  ptBR[3].x2 = n32X2 - 2; ptBR[3].y2 = n32Y2 - 1;
  ptFace->x = n32X1 + 2;
  ptFace->y = n32Y1 + 2;
  ptFace->width  = (n32X2 - n32X1) - 3;
  ptFace->height = (n32Y2 - n32Y1) - 3;

}
_vvoid _AdjustRectangle_(_vint32 &n32X1, _vint32 &n32Y1, _vint32 &n32X2, _vint32 &n32Y2,
                         _vint32 n32Width, _vint32 n32Height)
{
  // ..... Temporary variables
  _vint32 n32X, n32Y;
  
  // ..... Make sure that the rectangle begins from (0, 0)
  if(n32X1 != 0)
  {
    n32X2 -= n32X1;
    n32X1 = 0;
  }
  if(n32Y1 != 0)
  {
    n32Y2 -= n32Y1;
    n32Y1 = 0;
  }
  
  // ..... Make sure that the rectangle fits in the window's boundary
  if((n32X2 + 1) > (n32Width - 10))  n32X2 = n32Width - 10;
  if((n32Y2 + 1) > (n32Height - 10)) n32Y2 = n32Height - 10;
    
  // ..... Center rectangle
  n32X = (n32Width - (n32X2)) / 2;
  n32Y = (n32Height - (n32Y2)) / 2;
  n32X1 += n32X;
  n32Y1 += n32Y;
  n32X2 += n32X;
  n32Y2 += n32Y;
}
_vvoid _InitMsgBoxSize_(_EVortexMsgSize eSize, _vint32 &n32X1,
                        _vint32 &n32Y1, _vint32 &n32X2, _vint32 &n32Y2,
                        _vuint32 un32MaxWidth, _vuint32 un32MaxHeight)
{
  n32X1 = 0; n32Y1 = 0;
  if(eSize == ESizeTiny)
  { 
    n32X2 = 200; n32Y2 = 100; 
  }
  else if(eSize == ESizeSmall)
  { 
    n32X2 = 400; n32Y2 = 200; 
  }
  else if(eSize == ESizeMedium)
  { 
    n32X2 = 600; n32Y2 = 300; 
  }
  else if(eSize == ESizeLarge)
  { 
    n32X2 = 800; n32Y2 = 400; 
  }
  else
  { 
    n32X2 = 1000; n32Y2 = 500; 
  }
  _AdjustRectangle_(n32X1, n32Y1, n32X2, n32Y2, un32MaxWidth, un32MaxHeight);
}

// ============================================================================================
// End of vx11.cpp
// ============================================================================================
