// ############################################################################################
//  vimage.h
//  Definition & implementation  of _CVortexTexture, _CVortexAnimatedTexture, _CVortexImage,
//  _CVortexAnimatedImage, and _CVortexImageBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIMAGE_H_
#define _VIMAGE_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <stdlib.h>
// ***** VortexGE headers
#include <vvidenv.h>

// ============================================================================================
// Foward references of classes
// ============================================================================================
// ***** From "v3denv.h"
class _CVortex3DEnv;
  
// ============================================================================================
// Enumerations
// ============================================================================================
// ***** Address alignment hints
typedef enum _EVMAH_ { AAH_NEVER, AAH_SSE } _EVortexAAHint;

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The texture class
class _CVortexTexture
{
  protected:
    // Variables
    _vuint8  *mt_pun8Buffer;   // The texture buffer (in VRGB format)
    _vuint32 mt_un32Width;     // The texture width
    _vuint32 mt_un32Height;    // The texture height 
    _vuint32 mt_un32Mask;      // The address mask
    _vuint32 mt_un32MultShift; // The pre-multiply shift
    _vbool   mt_bAligned;      // Indicates that the buffer address is alignend
  
  public:
    // Constructor & destructor
    inline _CVortexTexture();
    inline ~_CVortexTexture();

  // Friend classes
  friend class _CVortexImageBuilder; 
  friend class _CVortexPanel3DBuffer;   
  friend class _CVortexPanel3D;   
};

// ***** The animated texture class
class _CVortexAnimatedTexture : public _CVortexTexture
{
  protected:
    // Variables
    _vuint8  *mt_pun8BufferArray; // The image buffer array (in VRGB format)
    _vuint8  *mt_pun8BufferMax;   // The maximum buffer address
    _vuint32 mt_un32FrameSize;    // Size of each frame in bytes
    _vuint32 mt_un32TotFrame;     // Total frames
    _vuint32 mt_un32CurFrame;     // Current frame
    _vuint32 mt_un32StepDelay;    // Stepping delay
    _vint8   mt_n8Step;           // The step factor
    _vbool   mt_bAligned;         // Indicates that the buffer address is alignend
   
  public:
    // Constructor & destructor
    inline _CVortexAnimatedTexture();
    inline ~_CVortexAnimatedTexture();
    // Image functions
    inline _vvoid NextFrame(_vbool bWarp = false);
    inline _vvoid PrevFrame(_vbool bWarp = false);
    inline _vvoid StepFrame(_vvoid);
    inline _vvoid StepFrame(_vuint32 un32Delay);

  // Friend classes
  friend class _CVortexImageBuilder; 
  friend class _CVortexPanel3DBuffer;   
  friend class _CVortexPanel3D;   
};

// ***** The image class
class _CVortexImage
{
  protected:
    // Variables
    _vuint8  *mt_pun8Buffer; // The image buffer (in VRGB format)
    _vuint32 mt_un32Width;   // The image width
    _vuint32 mt_un32Height;  // The image height 
    _vuint32 mt_un32SizeSL;  // The scanline size in bytes
    _vbool   mt_bAligned;    // Indicates that the buffer address is alignend
  
  public:
    // Constructor & destructor
    inline _CVortexImage();
    inline ~_CVortexImage();

  // Friend classes
  friend class _CVortexImageBuilder; 
  friend class _CVortex3DEnv;
  friend class _CVortexStaticSprite2D;
};

// ***** The animated image class
class _CVortexAnimatedImage : public _CVortexImage
{
  protected:
    // Variables
    _vuint8  *mt_pun8BufferArray; // The image buffer array (in VRGB format)
    _vuint8  *mt_pun8BufferMax;   // The maximum buffer address
    _vuint32 mt_un32FrameSize;    // Size of each frame in bytes
    _vuint32 mt_un32TotFrame;     // Total frames
    _vuint32 mt_un32CurFrame;     // Current frame
    _vuint32 mt_un32StepDelay;    // Stepping delay
    _vint8   mt_n8Step;           // The step factor
    _vbool   mt_bAligned;         // Indicates that the buffer address is alignend
   
  public:
    // Constructor & destructor
    inline _CVortexAnimatedImage();
    inline ~_CVortexAnimatedImage();
    // Image functions
    inline _vvoid NextFrame(_vbool bWarp = false);
    inline _vvoid PrevFrame(_vbool bWarp = false);
    inline _vvoid StepFrame(_vvoid);
    inline _vvoid StepFrame(_vuint32 un32Delay);

  // Friend classes
  friend class _CVortexImageBuilder; 
  friend class _CVortex3DEnv;
  friend class _CVortexStaticSprite2D;
  friend class _CVortexAImgControl;
};

// ***** The image builder class
class _CVortexImageBuilder
{
  protected:
    // Variables
    _TVortexUN8RGB *mt_ptBuffer;  // The image buffer (in triple RGB format)
    _vuint32       mt_un32Width;  // The image width
    _vuint32       mt_un32Height; // The image height
  
  public:
    // Constructor & destructor
    _CVortexImageBuilder();
    ~_CVortexImageBuilder();
    // Image functions
    inline _vuint32 GetWidth(_vvoid);
    inline _vuint32 GetHeight(_vvoid);
    inline _vvoid FreeImage(_vvoid);
    _vbool SmearEffect(_vuint8 un8Radius = 2);
    _vbool BlurEffect(_vuint8 un8Radius = 1);
    _vbool OilPaintingEffect(_vuint8 un8Factor = 1);
    _vbool AdjustBrightness(_vfloat fMultFactor, _vint8 un8AddFactor, 
                            _vbool bAffectRed = true, _vbool bAffectGreen = true, _vbool bAffectBlue = true);
    _vbool Grayscale(_vbool bUseLinearEquation = false);
    _vbool RoughResize(_vuint32 un32NewWidth, _vuint32 un32NewHeight);
    _vbool Resize(_vuint32 un32NewWidth, _vuint32 un32NewHeight);
    _vbool ResizeToFitBackground(_CVortex3DEnv *pcsV3DEnv = NULL);
    _vbool ResizeToFitTexture(_vuint32 un32TexWidth = 0, _vuint32 un32TexHeight = 0);
    _vbool TransferToImage(_CVortexImage &csVImage, _EVortexAAHint eVAAHint = AAH_SSE);
    _vbool TransferToAnimatedImage(_CVortexAnimatedImage &csVAnimatedImage, 
                                   _vuint8 un8SrcRowCount, _vuint8 un8SrcColCount,
                                   _EVortexAAHint eVAAHint = AAH_SSE);
    _vbool TransferToTexture(_CVortexTexture &csVTexture, _EVortexAAHint eVAAHint = AAH_SSE);
    _vbool TransferToAnimatedTexture(_CVortexAnimatedTexture &csVAnimatedTexture, 
                                     _vuint8 un8SrcRowCount, _vuint8 un8SrcColCount,
                                     _EVortexAAHint eVAAHint = AAH_SSE);
    // File I/O functions
    _vbool LoadBMP(const _vchar *pcFileName, _vbool bUseDithering = false);
    _vbool LoadVIB(const _vchar *pcFileName);
    _vbool SaveVIB(const _vchar *pcFileName);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The texture class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexTexture::_CVortexTexture()
{
  mt_bAligned      = false;
  mt_pun8Buffer    = NULL;
  mt_un32Width     = 0;
  mt_un32Height    = 0;
  mt_un32Mask      = 0;
  mt_un32MultShift = 0;
}
_CVortexTexture::~_CVortexTexture()
{
  if(mt_pun8Buffer)
  {
    if(mt_bAligned)
      ssefree(mt_pun8Buffer);
    else
      free(mt_pun8Buffer);
  }
}

// --------------------------------------------------------------------------------------------
// The animated texture class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexAnimatedTexture::_CVortexAnimatedTexture()
{
  mt_bAligned        = false;
  mt_pun8BufferArray = NULL;
  mt_pun8BufferMax   = NULL;
  mt_un32TotFrame    = 0;
  mt_un32CurFrame    = 0;
  mt_un32FrameSize   = 0;
  mt_un32StepDelay   = 0;
  mt_n8Step          = 0;
}
_CVortexAnimatedTexture::~_CVortexAnimatedTexture()
{
  mt_pun8Buffer = NULL;
  if(mt_pun8BufferArray)
  {
    if(mt_bAligned)
      ssefree(mt_pun8BufferArray);
    else
      free(mt_pun8BufferArray);
  }
}
_vvoid _CVortexAnimatedTexture::NextFrame(_vbool bWarp)
{ 
  if(mt_un32CurFrame < (mt_un32TotFrame - 1))
  {
    mt_un32CurFrame++;
    mt_pun8Buffer += mt_un32FrameSize;
  }
  else if(bWarp)
  {
    mt_un32CurFrame = 0;
    mt_pun8Buffer   = mt_pun8BufferArray;
  }
}
_vvoid _CVortexAnimatedTexture::PrevFrame(_vbool bWarp)
{
  if(mt_un32CurFrame > 0)
  {
    mt_un32CurFrame--;
    mt_pun8Buffer -= mt_un32FrameSize;
  }
  else if(bWarp)
  {
    mt_un32CurFrame = mt_un32TotFrame - 1;
    mt_pun8Buffer   = mt_pun8BufferMax;
  }
}
_vvoid _CVortexAnimatedTexture::StepFrame(_vvoid)
{
  // ..... Goto next frame
  if(mt_n8Step >= 1)
  {
    if(mt_un32CurFrame < (mt_un32TotFrame - 1))
    {
      mt_un32CurFrame++;
      mt_pun8Buffer += mt_un32FrameSize;
    }
    else
    {
      mt_n8Step = -1;
      mt_un32CurFrame--;
      mt_pun8Buffer -= mt_un32FrameSize;
    }
  }
  // ..... Goto previous frame
  else
  {
    if(mt_un32CurFrame > 0)
    {
      mt_un32CurFrame--;
      mt_pun8Buffer -= mt_un32FrameSize;
    }
    else
    {
      mt_n8Step = 1;
      mt_un32CurFrame++;
      mt_pun8Buffer += mt_un32FrameSize;
    }
  }
}
_vvoid _CVortexAnimatedTexture::StepFrame(_vuint32 un32Delay)
{
  // ..... Do delay
  mt_un32StepDelay++;
  if(mt_un32StepDelay < un32Delay) return;
  mt_un32StepDelay = 0;

  // ..... Do step
  StepFrame();
}

// --------------------------------------------------------------------------------------------
// The image class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexImage::_CVortexImage()
{
  mt_bAligned   = false;
  mt_pun8Buffer = NULL;
  mt_un32Width  = 0;
  mt_un32Height = 0;
  mt_un32SizeSL = 0;
}
_CVortexImage::~_CVortexImage()
{
  if(mt_pun8Buffer)
  {
    if(mt_bAligned)
      ssefree(mt_pun8Buffer);
    else
      free(mt_pun8Buffer);
  }
}

// --------------------------------------------------------------------------------------------
// The animated image class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexAnimatedImage::_CVortexAnimatedImage()
{
  mt_bAligned        = false;
  mt_pun8BufferArray = NULL;
  mt_pun8BufferMax   = NULL;
  mt_un32TotFrame    = 0;
  mt_un32CurFrame    = 0;
  mt_un32FrameSize   = 0;
  mt_un32StepDelay   = 0;
  mt_n8Step          = 0;
}
_CVortexAnimatedImage::~_CVortexAnimatedImage()
{
  mt_pun8Buffer = NULL;
  if(mt_pun8BufferArray)
  {
    if(mt_bAligned)
      ssefree(mt_pun8BufferArray);
    else
      free(mt_pun8BufferArray);
  }
}
_vvoid _CVortexAnimatedImage::NextFrame(_vbool bWarp)
{ 
  if(mt_un32CurFrame < (mt_un32TotFrame - 1))
  {
    mt_un32CurFrame++;
    mt_pun8Buffer += mt_un32FrameSize;
  }
  else if(bWarp)
  {
    mt_un32CurFrame = 0;
    mt_pun8Buffer   = mt_pun8BufferArray;
  }
}
_vvoid _CVortexAnimatedImage::PrevFrame(_vbool bWarp)
{
  if(mt_un32CurFrame > 0)
  {
    mt_un32CurFrame--;
    mt_pun8Buffer -= mt_un32FrameSize;
  }
  else if(bWarp)
  {
    mt_un32CurFrame = mt_un32TotFrame - 1;
    mt_pun8Buffer   = mt_pun8BufferMax;
  }
}
_vvoid _CVortexAnimatedImage::StepFrame(_vvoid)
{
  // ..... Goto next frame
  if(mt_n8Step >= 1)
  {
    if(mt_un32CurFrame < (mt_un32TotFrame - 1))
    {
      mt_un32CurFrame++;
      mt_pun8Buffer += mt_un32FrameSize;
    }
    else
    {
      mt_n8Step = -1;
      mt_un32CurFrame--;
      mt_pun8Buffer -= mt_un32FrameSize;
    }
  }
  // ..... Goto previous frame
  else
  {
    if(mt_un32CurFrame > 0)
    {
      mt_un32CurFrame--;
      mt_pun8Buffer -= mt_un32FrameSize;
    }
    else
    {
      mt_n8Step = 1;
      mt_un32CurFrame++;
      mt_pun8Buffer += mt_un32FrameSize;
    }
  }
}
_vvoid _CVortexAnimatedImage::StepFrame(_vuint32 un32Delay)
{
  // ..... Do delay
  mt_un32StepDelay++;
  if(mt_un32StepDelay < un32Delay) return;
  mt_un32StepDelay = 0;

  // ..... Do step
  StepFrame();
}

// --------------------------------------------------------------------------------------------
// The image builder class
// --------------------------------------------------------------------------------------------
// ***** Image functions
_vuint32 _CVortexImageBuilder::GetWidth(_vvoid)
{
  return(mt_un32Width);
}
_vuint32 _CVortexImageBuilder::GetHeight(_vvoid)
{
  return(mt_un32Height);
}
_vvoid _CVortexImageBuilder::FreeImage(_vvoid)
{
  if(mt_ptBuffer) free(mt_ptBuffer);
  mt_ptBuffer   = NULL;
  mt_un32Width  = 0;
  mt_un32Height = 0;
}

#endif
// ============================================================================================
// End of vimage.h
// ============================================================================================
