// ############################################################################################
//  vthread.cpp
//  Implementation of _CVortexThread class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <time.h>
#include <signal.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <vthread.h>

// ============================================================================================
// Static functions implementation
// ============================================================================================
// ***** Thread functions
_vvoid *_ThreadFunc_(_vvoid *pvData)
{
  (_PTVortexThreadData(pvData)->pcsThread)->mt_un64ExitStatus = (_PTVortexThreadData(pvData)->pcsThread->mt_pfnUserCallback)(_PTVortexThreadData(pvData));
  (_PTVortexThreadData(pvData)->pcsThread)->mt_bTEN = true;
  pthread_exit(NULL);
  return(NULL);
}

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The thread class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexThread::_CVortexThread()
{
  mt_tTD.pcsThread  = this;
  mt_tTD.pbExit     = &mt_bExit;
  mt_tTD.pbSuspend  = &mt_bSuspend;
  mt_tTD.pvUserData = NULL;

  mt_ptThreadHandle  = 0;
  mt_pfnUserCallback = NULL;
  mt_bExit           = false;
  mt_bSuspend        = true;
  mt_bTEN            = false;
  mt_un64ExitStatus  = 0;
}
_CVortexThread::~_CVortexThread()
{
  if(!Stop()) Kill();
}

// ***** Thread functions
_vbool _CVortexThread::Initialize(VThreadCallback pfnUserCallback, _vvoid *pvUserData)
{
  // ..... Check
  if(pfnUserCallback == NULL)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Invalid parameter");
    return(false);
  }
  
  // ..... Set variables
  mt_pfnUserCallback = pfnUserCallback;
  mt_tTD.pvUserData  = pvUserData;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexThread::Start(_vvoid)
{
  // ..... Temporary variables
  pthread_attr_t tType;
  sigset_t       tSig;
  _vint32        n32Dummy;
  _vint32        an32Sig[] = { SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGALRM, SIGTERM, 
                               SIGCHLD, SIGWINCH, SIGVTALRM, SIGPROF };
  
  // ..... Initialize attributes
  if(pthread_attr_init(&tType))
  {
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Could not initialize thread attributes");
  }
  pthread_attr_setdetachstate(&tType, PTHREAD_CREATE_DETACHED);
 
  // ..... Create thread
  if(pthread_create(&mt_ptThreadHandle, &tType, _ThreadFunc_, &mt_tTD))
  {
    VSetErrorCode(VERR_CreateThread);
    VError(_F_, _L_, "Could not create thread");
    return(false);
  }
  
  // ..... Mask asynchronous signals for this thread
  sigemptyset(&tSig);
  for(_vuint8 un8Cnt = 0; un8Cnt < 10; un8Cnt++)
    sigaddset(&tSig, an32Sig[un8Cnt]);
  pthread_sigmask(SIG_BLOCK, &tSig, 0);
    
  // ..... Allow thread to be asynchronously canceled
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &n32Dummy);

  // ..... Set variables
  mt_bSuspend = false;
  mt_bExit    = false;
  mt_bTEN     = false;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexThread::Stop(_vuint32 un32TimeOutmS)
{
  // ..... Temporary variables
  clock_t tC1, tC2;

  // ..... Check
  if(mt_ptThreadHandle == 0) return(true);

  // ..... Order the thread to exit itself and wait
  mt_bExit = true;
  tC1 = tC2 = clock();
  while(_vuint32(_vdouble(tC2 - tC1) / _vdouble(CLOCKS_PER_SEC) * _vdouble(1000)) < un32TimeOutmS)
  {
    mt_bExit = true;
    tC2 = clock();
    if(mt_bTEN)
    {
      mt_ptThreadHandle = 0;
      mt_bExit          = false;
      mt_bTEN           = false;
      return(true);
    }
  }

  // ..... Failed !
  mt_bExit = true;
  VSetErrorCode(VERR_NotResponding);
  VError(_F_, _L_, "The thread is not responding the exit command");
  return(false);
}
_vbool _CVortexThread::Kill(_vvoid)
{
  // ..... Check
  if(mt_ptThreadHandle == 0) return(true);

  // ..... Force stop thread
  if(pthread_cancel(mt_ptThreadHandle)) 
  {
    VWarning(_F_, _L_, "Thread termination using pthread_cancel() failed, trying pthread_kill()");
    if(pthread_kill(mt_ptThreadHandle, SIGKILL))
    {
      VWarning(_F_, _L_, "Thread termination using pthread_kill() also failed, thread is still running");
      return(false);
    }
  }
    
  // ..... Set variables
  mt_ptThreadHandle = 0;
  mt_bTEN           = false;

  // ..... OK !
  return(true);
}

// ============================================================================================
// End of vthread.cpp
// ============================================================================================
