// ############################################################################################
//  bench3.cpp
//  Benchmark various 3D operations
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ############################################################################################
// Header files
// ############################################################################################
// ***** C++ headers
#include <math.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// ***** VortexGE headers
#include <vbtypes.h>

// ############################################################################################
// Light intensity calculation
// ############################################################################################
inline _vfloat LightInten(const _vfloat afPanelPoint[4], const _vfloat afPanelNormal[4], const _vfloat afLightPos[4])
{
  // ..... Temporary variables
  _vfloat afT[4], fMag, fCos;
  
  // ..... Create the normalized light vector from the panel center to the omni light position
  afT[0] = afLightPos[0] - afPanelPoint[0];
  afT[1] = afLightPos[1] - afPanelPoint[1];
  afT[2] = afLightPos[2] - afPanelPoint[2];
  fMag = sqrt(afT[0] * afT[0] + afT[1] * afT[1] + afT[2] * afT[2]);
  afT[0] /= fMag;
  afT[1] /= fMag;
  afT[2] /= fMag;
    
  // ..... Return the angle between panel's normal and the light vector
  fCos = afPanelNormal[0] * afT[0] + afPanelNormal[1] * afT[1] + afPanelNormal[2] * afT[2];
  fCos = (fCos < 0) ? 0 : fCos;
  return(fCos);
}
inline _vfloat LightIntenSSE(const _vfloat afPanelPoint[4], const _vfloat afPanelNormal[4], const _vfloat afLightPos[4])
{
  static _vfloat _sse fTmp;
  _vuint32 ebx;
  __asm__ __volatile__
  (
    // esi -> panel's point    : P[? z y x] 
    // edi -> panel's normal   : N[? z y x] 
    // edx -> light's position : L[? z y x]    
    // ebx -> the cosine value
#if 0    
    // This is the original version :
    "movups (%%esi), %%xmm0         \n" // xmm0 = P[? z y x]                           (panel's point)
    "movups (%%edx), %%xmm1         \n" // xmm1 = L[? z y x]                           (light's position)
    "subps   %%xmm0, %%xmm1         \n" // xmm1 = L[? z y x] - P[? z y x] = V[? z y x] (the light vector)
    "movaps  %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x]
    "mulps   %%xmm0, %%xmm1         \n" // xmm1 = V[? z y x] ^ 2          = S[? z y x] (the squared V)
    "movaps  %%xmm1, %%xmm2         \n" // xmm2 = S[? z y x]
    "movaps  %%xmm1, %%xmm3         \n" // xmm3 = S[? z y x]
    "shufps   $0x55, %%xmm2, %%xmm2 \n" // xmm2 = S[? ? ? y]
    "shufps   $0xAA, %%xmm3, %%xmm3 \n" // xmm3 = S[? ? ? z]
    "addss   %%xmm2, %%xmm1         \n" // xmm1 = S[? ? ? x+y]
    "addss   %%xmm3, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z]
    "sqrtss  %%xmm1, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z] ^ 0.5    = M[? ? ? m] (the magnitude of V)
    "shufps   $0x00, %%xmm1, %%xmm1 \n" // xmm1 = M[? m m m]
    "divps   %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x] / M[? m m m] = R[? z y x] (the normalized V)
    "mulps  (%%edi), %%xmm0         \n" // xmm0 = R[? z y x] * N[? z y x] = C[? z y x] (the cosine component)
    "movaps  %%xmm0, %%xmm1         \n" // xmm1 = C[? z y x]
    "movaps  %%xmm0, %%xmm2         \n" // xmm2 = C[? z y x]
    "shufps   $0xFD, %%xmm1, %%xmm1 \n" // xmm1 = C[? ? ? y]
    "shufps   $0xFE, %%xmm2, %%xmm2 \n" // xmm2 = C[? ? ? z]
    "addss   %%xmm1, %%xmm0         \n" // xmm0 = C[? ? ? x+y]
    "addss   %%xmm2, %%xmm0         \n" // xmm0 = C[? ? ? x+y+z]          = I[? ? ? i] (the cosine value)
    "xorps   %%xmm4, %%xmm4         \n" // zero the xmm4
    "maxss   %%xmm4, %%xmm0         \n" // make sure the cosine value >= 0
    "movss   %%xmm0, (%%ebx)        \n" // save result
#else
    // This is the optimized version (runs slightly faster than the original version, with less acuracy) :
    "movups (%%esi), %%xmm0         \n" // xmm0 = P[? z y x]                           (panel's point)
    "movups (%%edx), %%xmm1         \n" // xmm1 = L[? z y x]                           (light's position)
    "subps   %%xmm0, %%xmm1         \n" // xmm1 = L[? z y x] - P[? z y x] = V[? z y x] (the light vector)
    "movaps  %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x]
    "mulps   %%xmm0, %%xmm1         \n" // xmm1 = V[? z y x] ^ 2          = S[? z y x] (the squared V)
    "movaps  %%xmm1, %%xmm2         \n" // xmm2 = S[? z y x]
    "shufps   $0x55, %%xmm2, %%xmm2 \n" // xmm2 = S[? ? ? y]
    "movaps  %%xmm1, %%xmm3         \n" // xmm3 = S[? z y x]
    "shufps   $0xAA, %%xmm3, %%xmm3 \n" // xmm3 = S[? ? ? z]
    "addss   %%xmm2, %%xmm1         \n" // xmm1 = S[? ? ? x+y]
    "addss   %%xmm3, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z]
    "rsqrtss %%xmm1, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z] ^ 0.5    = M[? ? ? m] (the magnitude of V)
    "shufps   $0x00, %%xmm1, %%xmm1 \n" // xmm1 = M[? m m m]
    "mulps   %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x] / M[? m m m] = R[? z y x] (the normalized V)
    "mulps  (%%edi), %%xmm0         \n" // xmm0 = R[? z y x] * N[? z y x] = C[? z y x] (the cosine component)
    "movaps  %%xmm0, %%xmm1         \n" // xmm1 = C[? z y x]
    "shufps   $0xFD, %%xmm1, %%xmm1 \n" // xmm1 = C[? ? ? y]
    "movaps  %%xmm0, %%xmm2         \n" // xmm2 = C[? z y x]
    "shufps   $0xFE, %%xmm2, %%xmm2 \n" // xmm2 = C[? ? ? z]
    "addss   %%xmm1, %%xmm0         \n" // xmm0 = C[? ? ? x+y]
    "addss   %%xmm2, %%xmm0         \n" // xmm0 = C[? ? ? x+y+z]          = I[? ? ? i] (the cosine value)
    "xorps   %%xmm4, %%xmm4         \n" // zero the xmm4
    "maxss   %%xmm4, %%xmm0         \n" // make sure the cosine value >= 0
    "movss   %%xmm0, (%%ebx)        \n" // save the final result
#endif    
    : "=b" (ebx)                                               \
    :  "0" (&fTmp), "S" (afPanelPoint), "D" (afPanelNormal), "d" (afLightPos) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "memory"
  );
  return(fTmp);
}

// ############################################################################################
// Light intensity calculation (RGB)
// ############################################################################################
inline _vvoid RGBLightInten(const _vfloat afPanelPoint[4], const _vfloat afPanelNormal[4], const _vfloat afLightPos[4], const _vuint8 aun8RGBSrc[3], _vuint8 aun8RGBDst[3])
{
  // ..... Temporary variables
  _vfloat afT[4], fMag, fCos;
  
  // ..... Create the normalized light vector from the panel center to the omni light position
  afT[0] = afLightPos[0] - afPanelPoint[0];
  afT[1] = afLightPos[1] - afPanelPoint[1];
  afT[2] = afLightPos[2] - afPanelPoint[2];
  fMag = sqrt(afT[0] * afT[0] + afT[1] * afT[1] + afT[2] * afT[2]);
  afT[0] /= fMag;
  afT[1] /= fMag;
  afT[2] /= fMag;
    
  // ..... Return the angle between panel's normal and the light vector
  fCos = afPanelNormal[0] * afT[0] + afPanelNormal[1] * afT[1] + afPanelNormal[2] * afT[2];
  fCos = (fCos < 0) ? 0 : fCos;
  aun8RGBDst[0] = _vuint8(aun8RGBSrc[0] * fCos);
  aun8RGBDst[1] = _vuint8(aun8RGBSrc[1] * fCos);
  aun8RGBDst[2] = _vuint8(aun8RGBSrc[2] * fCos);
}
inline _vvoid RGBLightIntenSSE(const _vfloat afPanelPoint[4], const _vfloat afPanelNormal[4], const _vfloat afLightPos[4], const _vuint8 aun8RGBSrc[3], _vuint8 aun8RGBDst[3])
{
  static _vfloat _sse fTmp;
  _vuint32 ebx;
  __asm__ __volatile__
  (
    // esi -> panel's point    : P[? z y x] 
    // edi -> panel's normal   : N[? z y x] 
    // edx -> light's position : L[? z y x]    
    // ebx -> the cosine value
    // This is the optimized version (runs slightly faster than the original version, with less acuracy) :
    "movups (%%esi), %%xmm0         \n" // xmm0 = P[? z y x]                           (panel's point)
    "movups (%%edx), %%xmm1         \n" // xmm1 = L[? z y x]                           (light's position)
    "subps   %%xmm0, %%xmm1         \n" // xmm1 = L[? z y x] - P[? z y x] = V[? z y x] (the light vector)
    "movaps  %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x]
    "mulps   %%xmm0, %%xmm1         \n" // xmm1 = V[? z y x] ^ 2          = S[? z y x] (the squared V)
    "movaps  %%xmm1, %%xmm2         \n" // xmm2 = S[? z y x]
    "shufps   $0x55, %%xmm2, %%xmm2 \n" // xmm2 = S[? ? ? y]
    "movaps  %%xmm1, %%xmm3         \n" // xmm3 = S[? z y x]
    "shufps   $0xAA, %%xmm3, %%xmm3 \n" // xmm3 = S[? ? ? z]
    "addss   %%xmm2, %%xmm1         \n" // xmm1 = S[? ? ? x+y]
    "addss   %%xmm3, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z]
    "rsqrtss %%xmm1, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z] ^ 0.5    = M[? ? ? m] (the magnitude of V)
    "shufps   $0x00, %%xmm1, %%xmm1 \n" // xmm1 = M[? m m m]
    "mulps   %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x] / M[? m m m] = R[? z y x] (the normalized V)
    "mulps  (%%edi), %%xmm0         \n" // xmm0 = R[? z y x] * N[? z y x] = C[? z y x] (the cosine component)
    "movaps  %%xmm0, %%xmm1         \n" // xmm1 = C[? z y x]
    "shufps   $0xFD, %%xmm1, %%xmm1 \n" // xmm1 = C[? ? ? y]
    "movaps  %%xmm0, %%xmm2         \n" // xmm2 = C[? z y x]
    "shufps   $0xFE, %%xmm2, %%xmm2 \n" // xmm2 = C[? ? ? z]
    "addss   %%xmm1, %%xmm0         \n" // xmm0 = C[? ? ? x+y]
    "addss   %%xmm2, %%xmm0         \n" // xmm0 = C[? ? ? x+y+z]          = I[? ? ? i] (the cosine value)
    "xorps   %%xmm4, %%xmm4         \n" // zero the xmm4
    "maxss   %%xmm4, %%xmm0         \n" // make sure the cosine value >= 0
    "movss   %%xmm0, (%%ebx)        \n" // save the final result
    : "=b" (ebx)                                               \
    :  "0" (&fTmp), "S" (afPanelPoint), "D" (afPanelNormal), "d" (afLightPos) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "memory"
  );
  aun8RGBDst[0] = _vuint8(aun8RGBSrc[0] * fTmp);
  aun8RGBDst[1] = _vuint8(aun8RGBSrc[1] * fTmp);
  aun8RGBDst[2] = _vuint8(aun8RGBSrc[2] * fTmp);
}
inline _vvoid RGBLightIntenSSEFull(const _vfloat afPanelPoint[4], const _vfloat afPanelNormal[4], const _vfloat afLightPos[4], const _vuint8 aun8RGBSrc[3], _vuint8 aun8RGBDst[3])
{
  
  static _vfloat _sse afTmp[4];
  _vuint32 ebx;
  
  afTmp[0] = _vfloat(aun8RGBSrc[0]);
  afTmp[1] = _vfloat(aun8RGBSrc[1]);
  afTmp[2] = _vfloat(aun8RGBSrc[2]);
  
  __asm__ __volatile__
  (
    // esi -> panel's point    : P[? z y x] 
    // edi -> panel's normal   : N[? z y x] 
    // edx -> light's position : L[? z y x]    
    // ebx -> the cosine value
    // This is the optimized version (runs slightly faster than the original version, with less acuracy) :
    "movups (%%esi), %%xmm0         \n" // xmm0 = P[? z y x]                           (panel's point)
    "movups (%%edx), %%xmm1         \n" // xmm1 = L[? z y x]                           (light's position)
    "subps   %%xmm0, %%xmm1         \n" // xmm1 = L[? z y x] - P[? z y x] = V[? z y x] (the light vector)
    "movaps  %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x]
    "mulps   %%xmm0, %%xmm1         \n" // xmm1 = V[? z y x] ^ 2          = S[? z y x] (the squared V)
    "movaps  %%xmm1, %%xmm2         \n" // xmm2 = S[? z y x]
    "shufps   $0x55, %%xmm2, %%xmm2 \n" // xmm2 = S[? ? ? y]
    "movaps  %%xmm1, %%xmm3         \n" // xmm3 = S[? z y x]
    "shufps   $0xAA, %%xmm3, %%xmm3 \n" // xmm3 = S[? ? ? z]
    "addss   %%xmm2, %%xmm1         \n" // xmm1 = S[? ? ? x+y]
    "addss   %%xmm3, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z]
    "rsqrtss %%xmm1, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z] ^ 0.5    = M[? ? ? m] (the magnitude of V)
    "shufps   $0x00, %%xmm1, %%xmm1 \n" // xmm1 = M[? m m m]
    "mulps   %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x] / M[? m m m] = R[? z y x] (the normalized V)
    "mulps  (%%edi), %%xmm0         \n" // xmm0 = R[? z y x] * N[? z y x] = C[? z y x] (the cosine component)
    "movaps  %%xmm0, %%xmm1         \n" // xmm1 = C[? z y x]
    "shufps   $0xFD, %%xmm1, %%xmm1 \n" // xmm1 = C[? ? ? y]
    "movaps  %%xmm0, %%xmm2         \n" // xmm2 = C[? z y x]
    "shufps   $0xFE, %%xmm2, %%xmm2 \n" // xmm2 = C[? ? ? z]
    "addss   %%xmm1, %%xmm0         \n" // xmm0 = C[? ? ? x+y]
    "addss   %%xmm2, %%xmm0         \n" // xmm0 = C[? ? ? x+y+z]          = I[? ? ? i] (the cosine value)
    "shufps   $0x00, %%xmm0, %%xmm0 \n" // xmm0 = I[? i i i]
    "xorps   %%xmm4, %%xmm4         \n" // zero the xmm4
    "maxps   %%xmm4, %%xmm0         \n" // make sure the cosine value >= 0
    "mulps  (%%ebx), %%xmm0         \n" // xmm0 = the RGB light intensity
    "movaps  %%xmm0, (%%ebx)        \n" // save the final result
    : "=b" (ebx)                                               \
    :  "0" (afTmp), "S" (afPanelPoint), "D" (afPanelNormal), "d" (afLightPos) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "memory"
  );
  aun8RGBDst[0] = _vuint8(afTmp[0]);
  aun8RGBDst[1] = _vuint8(afTmp[1]);
  aun8RGBDst[2] = _vuint8(afTmp[2]);
}

// ############################################################################################
// Logger
// ############################################################################################
inline _vvoid Log(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  _vchar  acBuff[1024];
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(acBuff, 1024, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
  printf(acBuff);
}

// ############################################################################################
// Timer
// ############################################################################################
_vuint64 GetCycles(_vvoid)
{
  _vuint64 un64Val;
  _vuint32 edi;
  __asm__ __volatile__
  (
    "rdtsc              \n"
    "mov %%eax,  (%%edi)\n"
    "mov %%edx, 4(%%edi)\n"
    : "=&D" (edi) : "0" (&un64Val) : "eax", "edx", "memory"
  );
  return(un64Val);
}

// ############################################################################################
// Just for testing
// ############################################################################################
_vvoid Test(_vvoid)
{
  _vfloat _sse afS[4] = {10, 11, 12, 13};
  _vfloat _sse afD[4] = {20, 21, 22, 23};
  _vfloat _sse afE[4] = { 0,  0,  0,  0};
  _vuint32 esi, edi;
  
  Log("Original         : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  __asm__ __volatile__
  (
    "xorps   %%xmm0, %%xmm0  \n"
    "movss  (%%esi), %%xmm0  \n"
    "movaps  %%xmm0, (%%edi) \n"
    : "=D" (edi)                                               \
    :  "0" (afE), "S" (afS)
    : "xmm0", "memory"
  );
  Log("movss            : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  Log("\n");

  Log("Original         : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  __asm__ __volatile__
  (
    "xorps   %%xmm0, %%xmm0         \n"
    "movss  (%%esi), %%xmm0         \n"
    "shufps   $0x3F, %%xmm0, %%xmm0 \n"
    "movaps  %%xmm0, (%%edi)        \n"
    : "=D" (edi)                                               \
    :  "0" (afE), "S" (afS)
    : "xmm0", "memory"
  );
  Log("shufps 3F (0333) : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  __asm__ __volatile__
  (
    "xorps   %%xmm0, %%xmm0         \n"
    "movss  (%%esi), %%xmm0         \n"
    "shufps   $0xCF, %%xmm0, %%xmm0 \n"
    "movaps  %%xmm0, (%%edi)        \n"
    : "=D" (edi)                                               \
    :  "0" (afE), "S" (afS)
    : "xmm0", "memory"
  );
  Log("shufps CF (3033) : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  __asm__ __volatile__
  (
    "xorps   %%xmm0, %%xmm0         \n"
    "movss  (%%esi), %%xmm0         \n"
    "shufps   $0xF3, %%xmm0, %%xmm0 \n"
    "movaps  %%xmm0, (%%edi)        \n"
    : "=D" (edi)                                               \
    :  "0" (afE), "S" (afS)
    : "xmm0", "memory"
  );
  Log("shufps F3 (3303) : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  __asm__ __volatile__
  (
    "xorps   %%xmm0, %%xmm0         \n"
    "movss  (%%esi), %%xmm0         \n"
    "shufps   $0xFC, %%xmm0, %%xmm0 \n"
    "movaps  %%xmm0, (%%edi)        \n"
    : "=D" (edi)                                               \
    :  "0" (afE), "S" (afS)
    : "xmm0", "memory"
  );
  Log("shufps FC (3330) : afE = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afE[0], afE[1], afE[2], afE[3]);
  Log("\n");
 
  Log("Original src     : afS = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afS[0], afS[1], afS[2], afS[3]);
  Log("Original dst     : afD = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afD[0], afD[1], afD[2], afD[3]);
  __asm__ __volatile__
  (
    "movaps (%%esi), %%xmm0         \n"
    "movaps (%%edi), %%xmm1         \n"
    "shufps   $0x4D, %%xmm0, %%xmm1 \n"
    "movaps  %%xmm0, (%%esi)        \n"
    "movaps  %%xmm1, (%%edi)        \n"
    : "=S" (esi), "=D" (edi)                                               \
    :  "0" (afS),  "1" (afD)
    : "xmm0", "xmm1", "memory"
  );
  Log("After shufps src : afS = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afS[0], afS[1], afS[2], afS[3]);
  Log("After shufps dst : afD = [%5.2f | %5.2f | %5.2f | %5.2f]\n", afD[0], afD[1], afD[2], afD[3]);
}

// ############################################################################################
// Main program (benchmaker)
// ############################################################################################
#define TOT_ITER   (1024 * 1024)
#define TOT_ITER_B (1024 * 1024 * 256)
#define MUL_FACT_B (TOT_ITER_B / TOT_ITER)
int main(void)
{
  // ***** Local variables
  _vuint32 un32Cnt;
  _vuint64 un64TS, un64TE, un64TC, un64TD;
  _vfloat  afSpd[2], fR1, fR2;
  
  _vuint8 aun8RGBSrc[3] = {31, 31, 31};
  _vuint8 aun8RGBDst[3];
  
  // ***** Aligned local variables
  _vfloat _sse afVec1[4]    = {  3,  2,  5, 0 };
  _vfloat _sse afVec2[4]    = {  0,  0,  1, 0 };
  _vfloat _sse afVec3[4]    = { 10, 20, 30, 0 };
  
  // ***** Display cycles / seconds
  un64TS = GetCycles(); sleep(0); un64TE = GetCycles();
  un64TD = un64TE - un64TS;
  un64TS = GetCycles(); sleep(1); un64TE = GetCycles();
  Log("\n%ull cycles per second\n", un64TE - un64TS - un64TD);
  
  // ***** Calculate overhead
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++);
  un64TE = GetCycles();
  un64TD = un64TE - un64TS;
  Log("\nLooped timing overhead = %ull cycles\n", un64TD);
  
  // ***** Just test
  Log("\n");
  Test();
  
  // ***** Benchmark emms and nop
  //       Mnemonics : nop     = 90
  //                   emms    = 0F77
  //                   any fpu = DD5C???? or D945??
  un64TC = 0;
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = GetCycles();
    un64TE = GetCycles();
    un64TC += (un64TE - un64TS);
  }
  Log("\nGet RDTSC overhead = %7.5f cycles\n", _vdouble(un64TC) / _vdouble(TOT_ITER));
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) __asm__ __volatile__ ("emms");
  un64TE = GetCycles();
  Log("emms               = %7.5f cycles\n", _vdouble(un64TE - un64TS - un64TD) / _vdouble(TOT_ITER));
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_B; un32Cnt++) __asm__ __volatile__ ("nop");
  un64TE = GetCycles();
  Log("nop                = %7.5f cycles\n", _vdouble(un64TE - un64TS - (un64TD * MUL_FACT_B)) / _vdouble(TOT_ITER_B));
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_B; un32Cnt++) __asm__ __volatile__ ("nop\nnop");
  un64TE = GetCycles();
  Log("nop nop            = %7.5f cycles\n", _vdouble(un64TE - un64TS - (un64TD * MUL_FACT_B)) / _vdouble(TOT_ITER_B));
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_B; un32Cnt++) __asm__ __volatile__ ("nop\nnop\nnop");
  un64TE = GetCycles();
  Log("nop nop nop        = %7.5f cycles\n", _vdouble(un64TE - un64TS - (un64TD * MUL_FACT_B)) / _vdouble(TOT_ITER_B));
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_B; un32Cnt++) __asm__ __volatile__ ("nop\nnop\nnop\nnop");
  un64TE = GetCycles();
  Log("nop nop nop nop    = %7.5f cycles\n", _vdouble(un64TE - un64TS - (un64TD * MUL_FACT_B)) / _vdouble(TOT_ITER_B));
  
  // ***** Benchmark light intensity calculation functions
  Log("\nBenchmarking LightIntensityCalculation functions ...\n");
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) fR1 = LightInten(afVec1, afVec2, afVec3);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - un64TD;
  afSpd[0] = _vdouble(un64TC) / _vdouble(TOT_ITER);
  Log("  C++ LS              : %8.5f : %7.5f cycles\n", fR1, afSpd[0]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) fR2 = LightIntenSSE(afVec1, afVec2, afVec3);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - un64TD;
  afSpd[1] = _vdouble(un64TC) / _vdouble(TOT_ITER);
  Log("  SSE LS              : %8.5f : %8.5f cycles (%4.2fX faster, %7.5f%%%% error)\n", fR2, afSpd[1], afSpd[0] / afSpd[1], fabs(fR1 - fR2) / fR1 * 100);
  
  // ***** Benchmark RGB light intensity calculation functions
  Log("\nBenchmarking LightIntensityCalculation functions ...\n");
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) RGBLightInten(afVec1, afVec2, afVec3, aun8RGBSrc, aun8RGBDst);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - un64TD;
  afSpd[0] = _vdouble(un64TC) / _vdouble(TOT_ITER);
  Log("  C++ LS RGB          : %2d %2d %2d : %7.5f cycles\n", aun8RGBDst[0], aun8RGBDst[1], aun8RGBDst[2], afSpd[0]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) RGBLightIntenSSE(afVec1, afVec2, afVec3, aun8RGBSrc, aun8RGBDst);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - un64TD;
  afSpd[1] = _vdouble(un64TC) / _vdouble(TOT_ITER);
  Log("  SSE RGB LS (/w C++) : %2d %2d %2d : %8.5f cycles (%4.2fX faster)\n", aun8RGBDst[0], aun8RGBDst[1], aun8RGBDst[2], afSpd[1], afSpd[0] / afSpd[1]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) RGBLightIntenSSEFull(afVec1, afVec2, afVec3, aun8RGBSrc, aun8RGBDst);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - un64TD;
  afSpd[1] = _vdouble(un64TC) / _vdouble(TOT_ITER);
  Log("  SSE RGB LS (full)   : %2d %2d %2d : %8.5f cycles (%4.2fX faster)\n", aun8RGBDst[0], aun8RGBDst[1], aun8RGBDst[2], afSpd[1], afSpd[0] / afSpd[1]);
  
  // ***** End
  return(0);
}

// ############################################################################################
// End of bench3.cpp
// ############################################################################################
